import doctest
import string
from pathlib import Path
from typing import Dict, List
from unicodedata import normalize
from unittest import TestCase

from pydantic import ValidationError

import everyvoice.text.utils
from everyvoice.config.text_config import Punctuation, Symbols, TextConfig
from everyvoice.model.feature_prediction.config import FeaturePredictionConfig
from everyvoice.tests.basic_test_case import BasicTestCase
from everyvoice.text.features import N_PHONOLOGICAL_FEATURES
from everyvoice.text.lookups import build_lookup, lookuptables_from_data
from everyvoice.text.phonemizer import AVAILABLE_G2P_ENGINES, get_g2p_engine
from everyvoice.text.text_processor import TextProcessor
from everyvoice.utils import (
    collapse_whitespace,
    generic_psv_filelist_reader,
    lower,
    nfc_normalize,
)


class TextTest(BasicTestCase):
    """Basic test for text input configuration"""

    def setUp(self) -> None:
        super().setUp()
        self.base_text_processor = TextProcessor(
            TextConfig(symbols=Symbols(letters=list(string.ascii_letters))),
        )

    def test_run_doctest(self):
        """Run doctests in everyvoice.utils"""
        results = doctest.testmod(everyvoice.text)
        self.assertFalse(results.failed, results)

    def test_text_to_sequence(self):
        text = "hello world"
        sequence = self.base_text_processor.encode_text(text)
        self.assertEqual(self.base_text_processor.decode_tokens(sequence, ""), text)

    def test_token_sequence_to_text(self):
        sequence = [51, 48, 55, 55, 58, 1, 66, 58, 61, 55, 47]
        self.assertEqual(self.base_text_processor.encode_text("hello world"), sequence)

    def test_hardcoded_symbols(self):
        self.assertEqual(
            self.base_text_processor.encode_text("\x80 "),
            [0, 1],
            "pad should be Unicode PAD symbol and index 0, whitespace should be index 1",
        )

    def test_cleaners_with_upper(self):
        text = "hello world"
        text_upper = "HELLO WORLD"
        upper_text_processor = TextProcessor(
            TextConfig(
                cleaners=[collapse_whitespace, lower],
                symbols=Symbols(letters=list(string.ascii_letters)),
            ),
        )
        sequence = upper_text_processor.encode_text(text_upper)
        self.assertEqual(upper_text_processor.decode_tokens(sequence, ""), text)

    def test_punctuation(self):
        text = "hello! How are you? My name's: foo;."
        upper_text_processor = TextProcessor(
            TextConfig(
                cleaners=[collapse_whitespace, lower],
                symbols=Symbols(letters=list(string.ascii_letters)),
            ),
        )
        tokens = upper_text_processor.apply_tokenization(
            upper_text_processor.normalize_text(text)
        )
        self.assertEqual(
            upper_text_processor.apply_punctuation_rules(tokens),
            [
                "h",
                "e",
                "l",
                "l",
                "o",
                "<EXCL>",
                " ",
                "h",
                "o",
                "w",
                " ",
                "a",
                "r",
                "e",
                " ",
                "y",
                "o",
                "u",
                "<QINT>",
                " ",
                "m",
                "y",
                " ",
                "n",
                "a",
                "m",
                "e",
                "<QUOTE>",
                "s",
                "<BB>",
                " ",
                "f",
                "o",
                "o",
                "<BB>",
                "<BB>",
            ],
        )

    def test_phonological_features(self):
        moh_config = FeaturePredictionConfig(
            contact=self.contact,
            text=TextConfig(
                cleaners=[collapse_whitespace, lower, nfc_normalize],
                symbols=Symbols(
                    letters=[
                        "ʌ̃̀ː",
                        "ʌ̃́ː",
                        "t͡ʃ",
                        "d͡ʒ",
                        "ʌ̃́",
                        "ʌ̃ː",
                        "kʰʷ",
                        "ũ̀ː",
                        "ɡʷ",
                        "áː",
                        "àː",
                        "aː",
                        "ʌ̃",
                        "èː",
                        "éː",
                        "iː",
                        "íː",
                        "ìː",
                        "kʷ",
                        "ṹː",
                        "óː",
                        "òː",
                        "ʃ",
                        "d",
                        "ɡ",
                        "á",
                        "a",
                        "é",
                        "e",
                        "í",
                        "i",
                        "k",
                        "n",
                        "ṹ",
                        "ũ",
                        "ó",
                        "o",
                        "r",
                        "h",
                        "t",
                        "s",
                        "w",
                        "f",
                        "j",
                        "ʔ",
                    ]
                ),
            ),
        )
        moh_text_processor = TextProcessor(moh_config.text)
        normalized_text = moh_text_processor.normalize_text("shéːkon")
        one_hot_tokens = moh_text_processor.encode_text(
            normalized_text, quiet=True
        )  # this finds ː as OOV
        g2p_tokens = moh_text_processor.encode_text(
            normalized_text, lang_id="moh", apply_g2p=True
        )
        feats = moh_text_processor.encode_text(
            normalized_text,
            lang_id="moh",
            apply_g2p=True,
            encode_as_phonological_features=True,
        )
        self.assertEqual(moh_text_processor.decode_tokens(g2p_tokens, ""), "séːɡũ")
        self.assertEqual(len(g2p_tokens), len(feats))
        self.assertNotEqual(len(g2p_tokens), len(one_hot_tokens))
        self.assertEqual(len(feats[0]), N_PHONOLOGICAL_FEATURES)

    def test_duplicates_removed(self):
        duplicate_symbols_text_processor = TextProcessor(
            TextConfig(
                symbols=Symbols(letters=list(string.ascii_letters), duplicate=["e"])
            )
        )
        self.assertEquals(
            len([x for x in duplicate_symbols_text_processor.symbols if x == "e"]), 1
        )

    def test_bad_symbol_configuration(self):
        with self.assertRaises(ValidationError):
            TextProcessor(
                TextConfig(symbols=Symbols(letters=list(string.ascii_letters), bad=[1]))
            )

    def test_dipgrahs(self):
        digraph_text_processor = TextProcessor(
            TextConfig(
                symbols=Symbols(letters=list(string.ascii_letters), digraph=["ee"])
            )
        )
        text = "ee"  # should be treated as "ee" and not two instances of "e"
        sequence = digraph_text_processor.encode_text(text)
        self.assertEqual(len(sequence), 1)

    def test_normalization(self):
        # This test doesn't really test very much, but just here to highlight that base cleaning doesn't involve NFC
        accented_text_processor = TextProcessor(
            TextConfig(
                cleaners=[nfc_normalize],
                symbols=Symbols(letters=list(string.ascii_letters), accented=["é"]),
            ),
        )
        text = "he\u0301llo world"
        sequence = accented_text_processor.encode_text(text)
        self.assertNotEqual(accented_text_processor.decode_tokens(sequence, ""), text)
        self.assertEqual(
            accented_text_processor.decode_tokens(sequence, ""),
            normalize("NFC", text),
        )
        self.assertNotEqual(
            self.base_text_processor.apply_cleaners(text), normalize("NFC", text)
        )

    def test_missing_symbol(self):
        text = "h3llo world"
        sequence = self.base_text_processor.encode_text(text)
        self.assertNotEqual(self.base_text_processor.decode_tokens(sequence), text)
        self.assertIn("3", self.base_text_processor.missing_symbols)
        self.assertEqual(self.base_text_processor.missing_symbols["3"], 1)


class LookupTableTest(TestCase):
    def test_build_lookup(self):
        """Make sure the original order of the keys is preserved"""
        key = "speaker"
        data = [
            {key: "Samuel"},
            {key: "Eric"},
            {key: "Eric"},
            {key: "Marc"},
            {key: "Aidan"},
            {key: "Marc"},
            {key: "Samuel"},
        ]
        speaker2id = build_lookup(data, key)
        self.assertDictEqual(
            speaker2id,
            {
                "Samuel": 0,
                "Eric": 1,
                "Marc": 2,
                "Aidan": 3,
            },
        )


class LookupTablesTest(TestCase):
    def test_lookuptables_from_data(self):
        """
        Text looluptables for a multilangual and multispeaker.
        """
        base_path = Path(__file__).parent / "data/lookuptable/"
        lang2id, speaker2id = lookuptables_from_data(
            (
                generic_psv_filelist_reader(base_path / "training_filelist.psv"),
                generic_psv_filelist_reader(base_path / "validation_filelist.psv"),
            )
        )
        self.assertDictEqual(
            lang2id, {"crk": 0, "git": 1, "str": 2}, "Language lookup tables differ"
        )
        self.assertDictEqual(
            speaker2id,
            {"0": 0, "1": 1, "2": 2, "3": 3},
            "Speaker lookup tables differ.",
        )

    def test_no_language(self):
        """
        Test a datasest that has no language.
        """

        def remove_language(data: List[Dict[str, str]]) -> List[Dict[str, str]]:
            for d in data:
                del d["language"]
            return data

        base_path = Path(__file__).parent / "data/lookuptable/"
        lang2id, speaker2id = lookuptables_from_data(
            (
                remove_language(
                    generic_psv_filelist_reader(base_path / "training_filelist.psv")
                ),
                remove_language(
                    generic_psv_filelist_reader(base_path / "validation_filelist.psv")
                ),
            )
        )
        self.assertDictEqual(lang2id, {}, "Language lookup tables differ")
        self.assertDictEqual(
            speaker2id,
            {"0": 0, "1": 1, "2": 2, "3": 3},
            "Speaker lookup tables differ.",
        )

    def test_no_speaker(self):
        """
        Test a datasest that has no speaker.
        """

        def remove_speaker(data: List[Dict[str, str]]) -> List[Dict[str, str]]:
            for d in data:
                del d["speaker"]
            return data

        base_path = Path(__file__).parent / "data/lookuptable/"
        lang2id, speaker2id = lookuptables_from_data(
            (
                remove_speaker(
                    generic_psv_filelist_reader(base_path / "training_filelist.psv")
                ),
                remove_speaker(
                    generic_psv_filelist_reader(base_path / "validation_filelist.psv")
                ),
            )
        )
        self.assertDictEqual(
            lang2id, {"crk": 0, "git": 1, "str": 2}, "Language lookup tables differ"
        )
        self.assertDictEqual(
            speaker2id,
            {},
            "Speaker lookup tables differ.",
        )


class TestG2p(TestCase):
    """Test G2P"""

    def test_many_available_langs(self):
        self.assertGreaterEqual(len(AVAILABLE_G2P_ENGINES), 20)

    def test_pua_chars(self):
        eng_g2p = get_g2p_engine("eng")
        und_g2p = get_g2p_engine("und")
        tokens = eng_g2p("h_e_l_l_o")
        self.assertEqual(
            tokens,
            ["e", "ɪ", "t", "ʃ", "_", "i", "_", "ɛ", "l", "_", "ɛ", "l", "_", "o", "ʊ"],
        )
        tokens = und_g2p("___")
        self.assertEqual(tokens, ["_", "_", "_"])

    def test_basic_g2p(self):
        eng_g2p = get_g2p_engine("eng")
        self.assertEqual(
            eng_g2p("hello world"), ["h", "ʌ", "l", "o", "ʊ", " ", "w", "ɜ˞", "l", "d"]
        )
        # keep's punctuation
        self.assertEqual(
            eng_g2p('hello "world"!!?.'),
            [
                "h",
                "ʌ",
                "l",
                "o",
                "ʊ",
                " ",
                '"',
                "w",
                "ɜ˞",
                "l",
                "d",
                '"',
                "!",
                "!",
                "?",
                ".",
            ],
        )
        # another language
        str_g2p = get_g2p_engine("str")
        self.assertEqual(str_g2p("SENĆOŦEN"), ["s", "ʌ", "n", "t͡ʃ", "ɑ", "θ", "ʌ", "n"])
        # test lang_id missing
        with self.assertRaises(NotImplementedError):
            get_g2p_engine("boop")

    def test_phonemizer_normalization(self):
        moh_g2p = get_g2p_engine("moh")
        self.assertEqual(moh_g2p("\u00E9"), ["\u00E9"])


class PunctuationTest(TestCase):
    def test_all(self):
        """Make sure we get the union of all punctuation characters when calling `all`."""
        punctuation = Punctuation()
        self.assertSetEqual(
            punctuation.all,
            {
                "?",
                "¿",
                "!",
                "¡",
                ",",
                ";",
                '"',
                "'",
                "«",
                "”",
                ":",
                "»",
                "-",
                "—",
                ".",
                "“",
                "…",
            },
        )


class SymbolsTest(TestCase):
    def test_all_except_punctuation(self):
        """Not withstanding the random new member variables defined by the
        user, we should get the union of them excluding what is in
        `punctuation`.
        """
        symbols = Symbols(
            dataset1=["a", "b"],
            dataset2=["X", "Y", "Z"],
        )
        self.assertSetEqual(
            symbols.all_except_punctuation, {"a", "b", "X", "Y", "Z", "<SIL>"}
        )



import typer
from merge_args import merge_args

from everyvoice.base_cli.interfaces import train_base_command_interface

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    help="End-to-end training: jointly train the FastSpeech2 and HiFiGAN networks",
)


@app.command()
@merge_args(train_base_command_interface)
def train(**kwargs):
    from everyvoice.base_cli.helpers import train_base_command
    from everyvoice.model.e2e.config import EveryVoiceConfig
    from everyvoice.model.e2e.dataset import E2EDataModule
    from everyvoice.model.e2e.model import EveryVoice

    train_base_command(
        model_config=EveryVoiceConfig,
        model=EveryVoice,
        data_module=E2EDataModule,
        monitor="validation/mel_spec_error",
        gradient_clip_val=None,
        **kwargs
    )



from pathlib import Path
from typing import Any, Sequence

import numpy as np
import torch
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.config import HiFiGANConfig
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.model import HiFiGAN
from everyvoice.text.text_processor import TextProcessor
from loguru import logger
from pympi import TextGrid
from pytorch_lightning.callbacks import Callback

from .config import FastSpeech2Config
from .type_definitions import SynthesizeOutputFormats


def get_synthesis_output_callbacks(
    output_type: Sequence[SynthesizeOutputFormats],
    output_dir: Path,
    config: FastSpeech2Config,
    output_key: str,
    device: torch.device,
    global_step: int,
    vocoder_model: HiFiGAN,
    vocoder_config: HiFiGANConfig,
    vocoder_global_step: int,
):
    """
    Given a list of desired output file formats, return the proper callbacks
    that will generate those files.
    """
    callbacks: list[Callback] = []
    if SynthesizeOutputFormats.spec in output_type:
        callbacks.append(
            PredictionWritingSpecCallback(
                config=config,
                global_step=global_step,
                output_dir=output_dir,
                output_key=output_key,
            )
        )
    if SynthesizeOutputFormats.textgrid in output_type:
        callbacks.append(
            PredictionWritingTextGridCallback(
                config=config,
                global_step=global_step,
                output_dir=output_dir,
                output_key=output_key,
            )
        )
    if SynthesizeOutputFormats.wav in output_type:
        callbacks.append(
            PredictionWritingWavCallback(
                config=config,
                device=device,
                global_step=global_step,
                output_dir=output_dir,
                output_key=output_key,
                vocoder_model=vocoder_model,
                vocoder_config=vocoder_config,
                vocoder_global_step=vocoder_global_step,
            )
        )

    return callbacks


class PredictionWritingCallbackBase(Callback):
    def __init__(
        self,
        file_extension: str,
        global_step: int,
        save_dir: Path,
    ) -> None:
        super().__init__()
        self.file_extension = file_extension
        self.global_step = f"ckpt={global_step}"
        self.save_dir = save_dir
        self.sep = "--"

        self.save_dir.mkdir(parents=True, exist_ok=True)

    def _get_filename(self, basename: str, speaker: str, language: str) -> Path:
        # We don't truncate or alter the filename here because the basename is
        # already truncated/cleaned in cli/synthesize.py
        path = self.save_dir / self.sep.join(
            [
                basename,
                speaker,
                language,
                self.global_step,
                self.file_extension,
            ]
        )
        path.parent.mkdir(
            parents=True, exist_ok=True
        )  # synthesizing spec allows nested outputs
        return path


class PredictionWritingSpecCallback(PredictionWritingCallbackBase):
    """
    This callback runs inference on a provided text-to-spec model and saves the resulting Mel spectrograms to disk as pytorch files. These can be used to fine-tune an EveryVoice spec-to-wav model.
    """

    def __init__(
        self,
        config: FastSpeech2Config,
        global_step: int,
        output_dir: Path,
        output_key: str,
    ):
        super().__init__(
            global_step=global_step,
            file_extension=f"spec-pred-{config.preprocessing.audio.input_sampling_rate}-{config.preprocessing.audio.spec_type}.pt",
            save_dir=output_dir / "synthesized_spec",
        )

        self.output_key = output_key
        self.config = config
        logger.info(f"Saving pytorch output to {self.save_dir}")

    def _get_filename(self, basename: str, speaker: str, language: str) -> Path:
        # We don't truncate or alter the filename here because the basename is
        # already truncated/cleaned in cli/synthesize.py
        # the spec should not have the global step printed because it is used to fine-tune
        # and the dataloader does not expect a global step in the filename
        path = self.save_dir / self.sep.join(
            [
                basename,
                speaker,
                language,
                self.file_extension,
            ]
        )
        path.parent.mkdir(
            parents=True, exist_ok=True
        )  # synthesizing spec allows nested outputs
        return path

    def on_predict_batch_end(  # pyright: ignore [reportIncompatibleMethodOverride]
        self,
        _trainer,
        _pl_module,
        outputs: dict[str, torch.Tensor | None],
        batch: dict[str, Any],
        _batch_idx: int,
        _dataloader_idx: int = 0,
    ):
        assert self.output_key in outputs and outputs[self.output_key] is not None
        assert "tgt_lens" in outputs and outputs["tgt_lens"] is not None
        for basename, speaker, language, data, unmasked_len in zip(
            batch["basename"],
            batch["speaker"],
            batch["language"],
            outputs[self.output_key],  # type: ignore
            outputs["tgt_lens"],
        ):
            torch.save(
                data[:unmasked_len].cpu(),
                self._get_filename(
                    basename=basename,
                    speaker=speaker,
                    language=language,
                ),
            )


class PredictionWritingTextGridCallback(PredictionWritingCallbackBase):
    """
    This callback runs inference on a provided text-to-spec model and saves the resulting textgrid of the predicted durations to disk. This can be used for evaluation.
    """

    def __init__(
        self,
        config: FastSpeech2Config,
        global_step: int,
        output_dir: Path,
        output_key: str,
    ):
        super().__init__(
            global_step=global_step,
            file_extension=f"{config.preprocessing.audio.input_sampling_rate}-{config.preprocessing.audio.spec_type}.TextGrid",
            save_dir=output_dir / "textgrids",
        )
        self.text_processor = TextProcessor(config.text)
        self.output_key = output_key
        self.config = config
        logger.info(f"Saving pytorch output to {self.save_dir}")

    def _get_filename(self, basename: str, speaker: str, language: str) -> Path:
        # We don't truncate or alter the filename here because the basename is
        # already truncated/cleaned in cli/synthesize.py
        # the textgrid should not have the global step printed because it is used to fine-tune
        # and the dataloader does not expect a global step in the filename
        path = self.save_dir / self.sep.join(
            [
                basename,
                speaker,
                language,
                self.file_extension,
            ]
        )
        path.parent.mkdir(
            parents=True, exist_ok=True
        )  # synthesizing spec allows nested outputs
        return path

    def frames_to_seconds(self, frames: int) -> float:
        return (
            frames * self.config.preprocessing.audio.fft_hop_size
        ) / self.config.preprocessing.audio.output_sampling_rate

    def on_predict_batch_end(  # pyright: ignore [reportIncompatibleMethodOverride]
        self,
        _trainer,
        _pl_module,
        outputs: dict[str, torch.Tensor | None],
        batch: dict[str, Any],
        _batch_idx: int,
        _dataloader_idx: int = 0,
    ):
        assert self.output_key in outputs and outputs[self.output_key] is not None
        assert (
            "duration_prediction" in outputs
            and outputs["duration_prediction"] is not None
        )
        for basename, speaker, language, raw_text, text, duration in zip(
            batch["basename"],
            batch["speaker"],
            batch["language"],
            batch["raw_text"],
            batch["text"],  # type: ignore
            outputs["duration_prediction"],
        ):
            # Get all durations in frames
            duration_frames = (
                torch.clamp(torch.round(torch.exp(duration) - 1), min=0).int().tolist()
            )
            # Get all input labels
            text_labels = self.text_processor.decode_tokens(
                text.tolist(), join_character=None
            )
            assert len(duration_frames) == len(
                text_labels
            ), f"can't synthesize {raw_text} because the number of predicted duration steps ({len(duration_frames)}) doesn't equal the number of input text labels ({len(text_labels)})"
            # get the duration of the audio: (sum_of_frames * hop_size) / sample_rate
            xmax_seconds = self.frames_to_seconds(sum(duration_frames))
            # create new textgrid
            new_tg = TextGrid(xmax=xmax_seconds)
            # create the tiers
            words: list[tuple[float, float, str]] = []
            phones: list[tuple[float, float, str]] = []
            raw_text_words = raw_text.split()
            current_word_duration = 0.0
            last_phone_end = 0.0
            last_word_end = 0.0
            phone_tier = new_tg.add_tier("phones")
            phone_annotation_tier = new_tg.add_tier("phone annotations")
            word_tier = new_tg.add_tier("words")
            word_annotation_tier = new_tg.add_tier("word annotations")
            # skip padding
            text_labels_no_padding = [tl for tl in text_labels if tl != "\x80"]
            duration_frames_no_padding = duration_frames[: len(text_labels_no_padding)]
            for label, duration in zip(
                text_labels_no_padding, duration_frames_no_padding
            ):
                # add phone label
                phone_duration = self.frames_to_seconds(duration)
                current_phone_end = last_phone_end + phone_duration
                interval = (last_phone_end, current_phone_end, label)
                phones.append(interval)
                phone_annotation_tier.add_interval(interval[0], interval[1], "")
                phone_tier.add_interval(*interval)
                last_phone_end = current_phone_end
                # accumulate phone to word label
                current_word_duration += phone_duration
                # if label is space or the last phone, add the word and recount
                if label == " " or len(phones) == len(text_labels_no_padding):
                    current_word_end = last_word_end + current_word_duration
                    interval = (
                        last_word_end,
                        current_word_end,
                        raw_text_words[len(words)],
                    )
                    words.append(interval)
                    word_tier.add_interval(*interval)
                    word_annotation_tier.add_interval(interval[0], interval[1], "")
                    last_word_end = current_word_end
                    current_word_duration = 0
            # get the filename
            filename = self._get_filename(
                basename=basename,
                speaker=speaker,
                language=language,
            )
            # write the file
            new_tg.to_file(filename)


class PredictionWritingWavCallback(PredictionWritingCallbackBase):
    """
    Given text-to-spec, this callback does spec-to-wav and writes wav files.
    """

    def __init__(
        self,
        output_dir: Path,
        config: FastSpeech2Config,
        output_key: str,
        device: torch.device,
        global_step: int,
        vocoder_model: HiFiGAN,
        vocoder_config: HiFiGANConfig,
        vocoder_global_step: int,
    ):
        super().__init__(
            file_extension="pred.wav",
            global_step=global_step,
            save_dir=output_dir / "wav",
        )

        self.output_key = output_key
        self.device = device
        self.config = config
        self.vocoder_model = vocoder_model
        self.vocoder_config = vocoder_config
        sampling_rate_change = (
            vocoder_config.preprocessing.audio.output_sampling_rate
            // vocoder_config.preprocessing.audio.input_sampling_rate
        )
        self.output_hop_size = (
            sampling_rate_change * vocoder_config.preprocessing.audio.fft_hop_size
        )
        self.file_extension = self.sep.join(
            (f"v_ckpt={vocoder_global_step}", self.file_extension)
        )

        logger.info(f"Saving wav output to {self.save_dir}")

    def synthesize_audio(self, outputs):
        from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.utils import (
            synthesize_data,
        )

        sr: int
        wavs: np.ndarray
        output_value = outputs[self.output_key]
        if output_value is not None:
            wavs, sr = synthesize_data(
                output_value, self.vocoder_model, self.vocoder_config
            )
        else:
            raise ValueError(
                f"{self.output_key} does not exist in the output of your model"
            )

        # wavs: [B (batch_size), T (samples)]
        assert (
            wavs.ndim == 2
        ), f"The generated audio contained more than 2 dimensions. First dimension should be B(atch) and the second dimension should be T(ime) in samples. Got {wavs.shape} instead."
        assert "output" in outputs and outputs["output"] is not None
        assert wavs.shape[0] == outputs["output"].size(
            0
        ), f"You provided {outputs['output'].size(0)} utterances, but {wavs.shape[0]} audio files were synthesized instead."

        # synthesize 16 bit audio
        # we don't do this higher up in the inference methods
        # because tensorboard logs require audio data as floats
        if (wavs >= -1.0).all() & (wavs <= 1.0).all():
            wavs = wavs * self.config.preprocessing.audio.max_wav_value
            wavs = wavs.astype("int16")
        return wavs, sr

    def on_predict_batch_end(  # pyright: ignore [reportIncompatibleMethodOverride]
        self,
        _trainer,
        _pl_module,
        outputs: dict[str, torch.Tensor | None],
        batch: dict[str, Any],
        _batch_idx: int,
        _dataloader_idx: int = 0,
    ):
        from scipy.io.wavfile import write

        logger.trace("Generating waveform...")

        wavs, sr = self.synthesize_audio(outputs)

        assert "tgt_lens" in outputs and outputs["tgt_lens"] is not None
        for basename, speaker, language, wav, unmasked_len in zip(
            batch["basename"],
            batch["speaker"],
            batch["language"],
            wavs,
            outputs["tgt_lens"],
        ):
            write(
                self._get_filename(
                    basename=basename,
                    speaker=speaker,
                    language=language,
                ),
                sr,
                # the vocoder output includes padding so we have to remove that
                wav[: (unmasked_len * self.output_hop_size)],
            )



import hashlib

from everyvoice.utils import slugify

BASENAME_MAX_LENGTH = 20


def truncate_basename(basename: str) -> str:
    """
    Shortens basename to BASENAME_MAX_LENGTH and uses the rest of basename to generate a sha1.
    This is done to make sure the file name stays short but that two utterances
    starting with the same prefix doesn't get ovverridden.
    """
    basename_cleaned = slugify(basename)
    if len(basename_cleaned) <= BASENAME_MAX_LENGTH:
        return basename_cleaned

    m = hashlib.sha1()
    m.update(bytes(basename, encoding="UTF-8"))
    return basename_cleaned[:BASENAME_MAX_LENGTH] + "-" + m.hexdigest()[:8]



from dataclasses import dataclass


@dataclass
class Frame:
    token_index: int
    time_index: int
    score: float


@dataclass
class Segment:
    label: str
    start: int
    end: int
    score: float

    def __repr__(self):
        return f"{self.label:2s} ({self.score:4.2f}): [{self.start:4d}, {self.end:4d})"

    def __len__(self):
        return self.end - self.start



"""
Common base class for the EveryVoice test suites

Adapted from https://github.com/ReadAlongs/Studio/blob/main/test/basic_test_case.py

"""

from pathlib import Path
from unittest import TestCase

from everyvoice.config.shared_types import ContactInformation


class BasicTestCase(TestCase):
    """A Basic Unittest build block class that comes bundled with
    the path to the test data (self.data_dir)

    For convenience, self.data_dir is pathlib.Path objects that can be used
    either with os.path functions or the shorter Path operators.
    E.g., these two lines are equivalent:
        text_file = os.path.join(self.data_dir, "ej-fra.txt")
        text_file = self.data_dir / "ej-fra.txt"
    """

    data_dir = Path(__file__).parent / "data"
    contact = ContactInformation(
        contact_name="Test Runner", contact_email="info@everyvoice.ca"
    )



# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Source: https://arxiv.org/abs/2108.10447

import torch
import torch.nn as nn
import torch.nn.functional as F


class AttentionCTCLoss(torch.nn.Module):
    def __init__(self, blank_logprob=-1):
        super(AttentionCTCLoss, self).__init__()
        self.log_softmax = torch.nn.LogSoftmax(dim=-1)
        self.blank_logprob = blank_logprob
        self.CTCLoss = nn.CTCLoss(zero_infinity=True)

    def forward(self, attn_logprob, in_lens, out_lens):
        key_lens = in_lens
        query_lens = out_lens
        max_key_len = attn_logprob.size(-1)

        # Reorder input to [query_len, batch_size, key_len]
        attn_logprob = attn_logprob.squeeze(1)
        attn_logprob = attn_logprob.permute(1, 0, 2)

        # Add blank label
        attn_logprob = F.pad(
            input=attn_logprob, pad=(1, 0, 0, 0, 0, 0), value=self.blank_logprob
        )

        # Convert to log probabilities
        # Note: Mask out probs beyond key_len
        key_inds = torch.arange(
            max_key_len + 1, device=attn_logprob.device, dtype=torch.long
        )
        attn_logprob.masked_fill_(
            key_inds.view(1, 1, -1) > key_lens.view(1, -1, 1),  # key_inds >= key_lens+1
            -1e15,
        )
        attn_logprob = self.log_softmax(attn_logprob)

        # Target sequences
        target_seqs = key_inds[1:].unsqueeze(0)
        target_seqs = target_seqs.repeat(key_lens.numel(), 1)

        # Evaluate CTC loss
        cost = self.CTCLoss(
            attn_logprob, target_seqs, input_lengths=query_lens, target_lengths=key_lens
        )
        return cost


class AttentionBinarizationLoss(torch.nn.Module):
    def __init__(self):
        super(AttentionBinarizationLoss, self).__init__()

    def forward(self, hard_attention, soft_attention, eps=1e-12):
        log_sum = torch.log(
            torch.clamp(soft_attention[hard_attention == 1], min=eps)
        ).sum()
        return -log_sum / hard_attention.sum()



import csv
import json
import os
import re
import sys
from contextlib import contextmanager
from datetime import datetime
from functools import partial
from itertools import islice
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional
from unicodedata import normalize

import yaml
from loguru import logger

from everyvoice import exceptions
from everyvoice.config.type_definitions import TargetTrainingTextRepresentationLevel

# Regular expression matching whitespace:
_whitespace_re = re.compile(r"\s+")
# Regular expression matching non-slug characters:
special_chars = re.compile(r"[\W]+")


def slugify(
    text: str, repl: str = "-", limit_to_n_characters: Optional[int] = None
) -> str:
    """Create a slugified representation of input text
       (i.e. without special characters) and optionally
       limited to a specific number of characters.

       >>> slugify('gya?a')
       'gya-a'

       >>> slugify('gya?a', repl='')
       'gyaa'

       >>> slugify('gya?a', limit_to_n_characters=3)
       'gya'

       >>> slugify('!@#$%^&*(){}<>? :;|+=/\\'"', limit_to_n_characters=3)
       '-'

       >>> slugify('!@#$%^&*(){}<>? :;|+=/\\'"', repl='', limit_to_n_characters=3)
       ''

    Args:
        text (str): text to slugify
        repl (str): character to replace special character with. defaults to -
        limit_to_n_characters (Optional[int]): return first n characters of output

    Returns:
        str: slugified string
    """
    slugified_text = re.sub(special_chars, repl, text)

    if limit_to_n_characters is None:
        return slugified_text
    else:
        return slugified_text[:limit_to_n_characters]


def filter_dataset_based_on_target_text_representation_level(
    target_text_representation_level: TargetTrainingTextRepresentationLevel,
    dataset: list[dict],
    name: str,
    batch_size: int,
) -> list[dict]:
    # remove dataset samples that don't exist for the target training representation
    match target_text_representation_level:
        case TargetTrainingTextRepresentationLevel.characters:
            target_training_text_key = "character_tokens"
        case (  # noqa: E211
            TargetTrainingTextRepresentationLevel.ipa_phones
            | TargetTrainingTextRepresentationLevel.phonological_features
        ):
            target_training_text_key = "phone_tokens"
        case _:
            raise NotImplementedError(
                f"{target_text_representation_level} have not yet been implemented."
            )

    filtered_dataset = [
        item
        for item in dataset
        if target_training_text_key in item and item[target_training_text_key]
    ]
    n_filtered_items = len(dataset) - len(filtered_dataset)
    if n_filtered_items:
        logger.warning(
            f"Removing {n_filtered_items} from your {name} set because they do not have text values for the target training representation level {target_text_representation_level}. Either change the target training representation level or update the information in your data and re-run preprocessing if you want to use this data."
        )
        dataset = filtered_dataset
    if batch_size > len(filtered_dataset):
        logger.error(
            f"Sorry you do not have enough {target_text_representation_level} data in your current {name} filelist to run the model with a batch size of {batch_size}."
        )
        sys.exit(1)
    return dataset


def check_dataset_size(batch_size: int, number_of_samples: int, name: str):
    if batch_size > number_of_samples:
        reduce_train_split = (
            "You can also decrease the train split to increase the number of samples in your validation dataset."
            if "val" in name
            else ""
        )
        logger.error(
            f"Your {name} dataset only has {number_of_samples} samples, but you have a defined batch size of {batch_size}. Please either add more data or decrease your batch size. {reduce_train_split}"
        )
        sys.exit(1)


def return_configs_from_dir(dir: Path) -> Dict[str, Path]:
    return {os.path.basename(path)[:-5]: path for path in dir.glob("*.yaml")}


def get_current_time():
    return str(int(datetime.now().timestamp()))


def _flatten(structure, key="", path="", flattened=None):
    """
    >>> _flatten({"a": {"b": 2, "c": {"d": "e"}, "f": 4}, "g": 5})
    {'a_b': 2, 'a_c_d': 'e', 'a_f': 4, 'g': 5}
    """
    if flattened is None:
        flattened = {}
    if not isinstance(structure, dict):
        flattened[(f"{path}_" if path else "") + key] = structure
    else:
        for new_key, value in structure.items():
            _flatten(value, new_key, (f"{path}_" if path else "") + key, flattened)
    return flattened


def load_config_from_json_or_yaml_path(path: Path):
    if not path.exists():
        raise ValueError(f"Config file '{path}' does not exist")
    with open(path, "r", encoding="utf8") as f:
        config = json.load(f) if path.suffix == ".json" else yaml.safe_load(f)
    if not config:
        raise exceptions.InvalidConfiguration(f"Your configuration at {path} was empty")
    return config


def expand_config_string_syntax(config_arg: str) -> dict:
    """Expand a string of the form "key1=value1" into a dict."""
    config_dict: Any = {}
    try:
        key, value = config_arg.split("=")
    except ValueError as e:
        raise ValueError(f"Invalid config string: {config_arg} - missing '='") from e
    current_dict = config_dict
    keys = key.split(".")
    for key in keys[:-1]:
        current_dict[key] = {}
        current_dict = current_dict[key]
    current_dict[keys[-1]] = value
    return config_dict


def update_config_from_cli_args(arg_list: List[str], original_config):
    if arg_list is None or not arg_list:
        return original_config
    for arg in arg_list:
        key, value = arg.split("=")
        logger.info(f"Updating config '{key}' to value '{value}'")
        original_config = original_config.update_config(
            expand_config_string_syntax(arg)
        )
    return original_config


def original_hifigan_leaky_relu(x):
    import torch.nn.functional as F

    return F.leaky_relu(x, 0.1)


def plot_spectrogram(spectrogram):
    import matplotlib.pylab as plt

    fig, ax = plt.subplots(figsize=(10, 2))
    im = ax.imshow(spectrogram, aspect="auto", origin="lower", interpolation="none")
    plt.colorbar(im, ax=ax)

    fig.canvas.draw()
    plt.close()

    return fig


def write_filelist(files, path):

    with open(path, "w", encoding="utf8") as f:
        if not files:
            logger.warning(f"Writing empty filelist file {path}")
            print("", file=f)  # header line, empty because we don't know the fields
            return
        # The base set of expected fieldnames
        fieldnames = [
            "basename",
            "language",
            "speaker",
            "characters",
            "character_tokens",
            "phones",
            "phone_tokens",
        ]
        # The fieldnames we actually found
        found_fieldnames = sorted(files[0].keys())
        # Use fieldnames in the order we expect, and append unexpected ones to the end
        fieldnames = [x for x in fieldnames if x in found_fieldnames] + [
            x for x in found_fieldnames if x not in fieldnames
        ]
        writer = csv.DictWriter(
            f,
            fieldnames=fieldnames,
            delimiter="|",
            quoting=csv.QUOTE_NONE,
            escapechar="\\",
        )
        writer.writeheader()
        for f in files:
            writer.writerow(f)


def lower(text):
    """
    >>> lower("MiXeD ÇÀSÉ")
    'mixed çàsé'
    """
    return text.lower()


def nfc_normalize(text):
    """
    >>> nfc_normalize("éçà")
    'éçà'
    """
    return normalize("NFC", text)


def read_festival(
    path,
    record_limit: int = 0,  # if non-zero, read only this many records
    text_field_name: str = "text",
):
    """Read Festival format into filelist
    Args:
        path (Path): Path to festival format filelist
        record_limit: if non-zero, read only that many records
        text_field_name (str): the keyname for the returned text. Default is 'text'.
    Raises:
        ValueError: the file is not valid festival input
    """
    festival_pattern = re.compile(
        r"""
    \(\s*
    (?P<basename>[\w\d\-\_.]*)
    \s*
    "(?P<text>[^"]*)"
    \s*\)
    """,
        re.VERBOSE,
    )
    data = []
    f: Iterable[str]
    with open(path, encoding="utf-8") as f:
        if record_limit:
            f = islice(f, record_limit)
        for line in f:
            if match := re.search(festival_pattern, line.strip()):
                basename = match["basename"].strip()
                text = match["text"].strip()
                data.append({"basename": basename, text_field_name: text})
            else:
                raise ValueError(f'File {path} is not in the "festival" format.')
    return data


def sniff_and_return_filelist_data(path):
    """Sniff csv, and return dialect if not festival format:
    ( LJ0002 "this is the festival format" )
    Args:
        path (Path): path to filelist
    Returns:
        False if not csv
    """
    festival_pattern = re.compile(r'\( [\w\d_]* "[^"]*" \)')
    with open(path, newline="", encoding="utf8") as f:
        data = f.read(1024)
        f.seek(0)
        if re.search(festival_pattern, data):
            return read_festival(path)
        else:
            dialect = csv.Sniffer().sniff(data)
            reader = csv.DictReader(f, dialect=dialect)
            return list(reader)


def generic_dict_loader(
    path: str | os.PathLike,
    delimiter="|",
    quoting=csv.QUOTE_NONE,
    escapechar="\\",
    fieldnames=None,
    file_has_header_line=True,
    record_limit: int = 0,
) -> list[dict]:
    """This is the base function that should be used to parse *sv style tabular filelists.
        The psv variant can also be imported with generic_psv_filelist_reader
        The csv variant can be imported with generic_csv_filelist_reader

    Args:
        path (_type_): path to *sv tabular filelist
        delimiter (str, optional): column delimiter. Defaults to "|".
        quoting (_type_, optional): quoting strategy. Defaults to csv.QUOTE_NONE.
        escapechar (str, optional): escape character. Defaults to "\".
        fieldnames (_type_, optional): fieldnames to parse. Defaults to None.
        file_has_header_line (bool, optional): whether file has header line. Defaults to True.
        record_limit (int): if non-zero, read only this many records. Defaults to 0.

    Returns:
        list[dict]: a list of dicts representing the rows in the filelist
    """
    assert fieldnames is not None or file_has_header_line
    f: Iterable[str]
    with open(path, "r", newline="", encoding="utf8") as f:
        if record_limit:
            f = islice(f, record_limit)
        reader = csv.DictReader(
            f,
            fieldnames=fieldnames,
            delimiter=delimiter,
            quoting=quoting,
            escapechar=escapechar,
        )
        # When fieldnames is given, csv.DictReader assumes the first line is a data
        # line.  Skip it if the file has a header line.
        if fieldnames and file_has_header_line:
            next(reader)
        files = []
        for file in list(reader):
            if "basename" in file:
                file["basename"] = os.path.splitext(file["basename"])[0]
            files.append(file)
    return files


generic_psv_filelist_reader = generic_dict_loader
generic_xsv_filelist_reader = generic_dict_loader
generic_csv_filelist_reader = partial(generic_dict_loader, delimiter=",")


def collapse_whitespace(text):
    """
    >>> collapse_whitespace("  asdf  	   qwer   ")
    ' asdf qwer '
    """
    return re.sub(_whitespace_re, " ", text)


@contextmanager
def tqdm_joblib_context(tqdm_instance):
    """Context manager to make tqdm compatible with joblib.Parallel

    Runs the parallel jobs using joblib, but displays the nicer tqdm progress bar
    Only tested with tqdm.tqdm, but should also work with tqdm.notepad.tqdm and
    other variants

    Usage:
        with tqdm_joblib_context(tqdm(desc="my description", total=len(job_list))):
            joblib.Parallel(n_jobs=cpus)(delayed(fn)(item) for item in job_list)
    """
    import joblib.parallel

    class ParallelCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, out):
            tqdm_instance.update(n=self.batch_size)
            super().__call__(out)

    old_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = ParallelCallback
    try:
        yield
    finally:
        tqdm_instance.close()
        joblib.parallel.BatchCompletionCallBack = old_callback


def n_times(n: int) -> str:
    """Return a grammatically correct version of n times for n > 0.

    >>> n_times(1)
    'once'
    >>> n_times(2)
    'twice'
    >>> n_times(1001)
    '1001 times'
    """
    if n == 1:
        return "once"
    if n == 2:
        return "twice"
    return f"{n} times"



# *****************************************************************************
#  Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#      * Neither the name of the NVIDIA CORPORATION nor the
#        names of its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL NVIDIA CORPORATION BE LIABLE FOR ANY
#  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# *****************************************************************************
import functools

import numpy as np
from scipy import ndimage
from scipy.stats import betabinom


class BetaBinomialInterpolator:
    """Interpolates alignment prior matrices to save computation.
    Calculating beta-binomial priors is costly. Instead cache popular sizes
    and use img interpolation to get priors faster.
    """

    def __init__(self, round_mel_len_to=100, round_text_len_to=20):
        self.round_mel_len_to = round_mel_len_to
        self.round_text_len_to = round_text_len_to
        self.bank = functools.lru_cache(beta_binomial_prior_distribution)

    def round(self, val, to):
        return max(1, int(np.round((val + 1) / to))) * to

    def __call__(self, w, h):
        bw = self.round(w, to=self.round_mel_len_to)
        bh = self.round(h, to=self.round_text_len_to)
        ret = ndimage.zoom(self.bank(bw, bh).T, zoom=(w / bw, h / bh), order=1)
        assert ret.shape[0] == w, ret.shape
        assert ret.shape[1] == h, ret.shape
        return ret


def beta_binomial_prior_distribution(phoneme_count, mel_count, scaling=1.0):
    P = phoneme_count
    M = mel_count
    x = np.arange(0, P)
    mel_text_probs = []
    for i in range(1, M + 1):
        a, b = scaling * i, scaling * (M + 1 - i)
        rv = betabinom(P, a, b)
        mel_i_prob = rv.pmf(x)
        mel_text_probs.append(mel_i_prob)
    return np.array(mel_text_probs)



from pathlib import Path

import torch
import torchaudio
from everyvoice.dataloader import BaseDataModule
from everyvoice.model.vocoder.config import VocoderConfig
from everyvoice.utils import check_dataset_size
from torch.utils.data import Dataset

from .utils import get_all_segments


class SpecDataset(Dataset):
    def __init__(
        self, audio_files, config: VocoderConfig, use_segments=False, finetune=False
    ):
        self.config = config
        self.sep = "--"
        self.use_segments = use_segments
        self.audio_files = audio_files
        self.preprocessed_dir = Path(self.config.preprocessing.save_dir)
        self.finetune = self.config.training.finetune
        self.segment_size = self.config.preprocessing.audio.vocoder_segment_size
        self.output_sampling_rate = self.config.preprocessing.audio.output_sampling_rate
        self.input_sampling_rate = self.config.preprocessing.audio.input_sampling_rate
        self.sampling_rate_change = (
            self.output_sampling_rate // self.input_sampling_rate
        )
        self.input_hop_size = self.config.preprocessing.audio.fft_hop_size
        self.output_hop_size = (
            self.config.preprocessing.audio.fft_hop_size * self.sampling_rate_change
        )

    def __getitem__(self, index):
        """
        x = mel spectrogram from potentially downsampled audio or from acoustic feature prediction
        y = waveform from potentially upsampled audio
        y_mel = mel spectrogram calculated from y
        """
        item = self.audio_files[index]
        speaker = "default" if "speaker" not in item else item["speaker"]
        language = "default" if "language" not in item else item["language"]
        y, _ = torchaudio.load(
            (
                self.config.preprocessing.save_dir
                / "audio"
                / self.sep.join(
                    [
                        item["basename"],
                        speaker,
                        language,
                        f"audio-{self.output_sampling_rate}.wav",
                    ]
                )
            ).__str__()
        )
        y = (
            y.squeeze()
        )  # [samples] should be output sample rate, squeeze to get rid of channels
        y_mel = torch.load(
            self.preprocessed_dir
            / "spec"
            / self.sep.join(
                [
                    item["basename"],
                    speaker,
                    language,
                    f"spec-{self.output_sampling_rate}-{self.config.preprocessing.audio.spec_type}.pt",
                ]
            )
        )  # [mel_bins, frames]
        if self.finetune:
            # If finetuning, use the synthesized spectral features
            x = torch.load(
                self.preprocessed_dir
                / "synthesized_spec"
                / self.sep.join(
                    [
                        item["basename"],
                        speaker,
                        language,
                        f"spec-pred-{self.input_sampling_rate}-{self.config.preprocessing.audio.spec_type}.pt",
                    ]
                )
            ).transpose(0, 1)
        else:
            x = torch.load(
                self.preprocessed_dir
                / "spec"
                / self.sep.join(
                    [
                        item["basename"],
                        speaker,
                        language,
                        f"spec-{self.input_sampling_rate}-{self.config.preprocessing.audio.spec_type}.pt",
                    ]
                )
            )  # [mel_bins, frames]
        if self.use_segments:
            x, y, y_mel = get_all_segments(
                x, y, y_mel, self.segment_size, self.output_hop_size
            )

        return (x, y, self.audio_files[index]["basename"], y_mel)

    def __len__(self):
        return len(self.audio_files)

    def get_labels(self):
        return [x["label"] for x in self.audio_files]


class HiFiGANDataModule(BaseDataModule):
    def __init__(self, config: VocoderConfig):
        super().__init__(config=config)
        self.use_weighted_sampler = config.training.use_weighted_sampler
        self.batch_size = config.training.batch_size

    def load_dataset(self):
        self.train_dataset = self.config.training.filelist_loader(
            self.config.training.training_filelist
        )
        self.val_dataset = self.config.training.filelist_loader(
            self.config.training.validation_filelist
        )

    def prepare_data(self):
        self.load_dataset()
        train_samples = len(self.train_dataset)
        val_samples = len(self.val_dataset)
        check_dataset_size(self.batch_size, train_samples, "training")
        check_dataset_size(self.batch_size, val_samples, "validation")
        self.train_dataset = SpecDataset(
            self.train_dataset, self.config, use_segments=True
        )
        self.val_dataset = SpecDataset(
            self.val_dataset, self.config, use_segments=False
        )
        # save it to disk
        torch.save(self.train_dataset, self.train_path)
        torch.save(self.val_dataset, self.val_path)


class HiFiGANFineTuneDataModule(BaseDataModule):
    def __init__(self, config: VocoderConfig):
        super().__init__(config=config)
        self.use_weighted_sampler = config.training.use_weighted_sampler
        self.batch_size = config.training.batch_size

    def load_dataset(self):
        self.dataset = SpecDataset(config=self.config, finetune=True)



import os
from pathlib import Path

import typer

from .utils import (
    TextHash,
    create_text_grid_from_segments,
    create_transducer,
    read_text,
)

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    help="An alignment tool based on CTC segmentation to split long audio into utterances",
)


def complete_path():
    return []


@app.command()
def align_single(
    text_path: Path = typer.Argument(
        ..., exists=True, file_okay=True, dir_okay=False, autocompletion=complete_path
    ),
    audio_path: Path = typer.Argument(
        ..., exists=True, file_okay=True, dir_okay=False, autocompletion=complete_path
    ),
    sample_rate: int = typer.Option(
        16000, help="The target sample rate for the model."
    ),
    word_padding: int = typer.Option(0, help="How many frames to pad around words."),
    sentence_padding: int = typer.Option(
        0, help="How many frames to pad around sentences (additive with word-padding)."
    ),
    debug: bool = typer.Option(False, help="Print debug statements"),
):
    print("loading model...")
    import torch
    import torchaudio

    from .heavy import load_model

    model, labels = load_model()
    wav, sr = torchaudio.load(audio_path)
    if sr != sample_rate:
        print(f"resampling audio from {sr} to {sample_rate}")
        wav = torchaudio.functional.resample(wav, sr, sample_rate)
        fn, ext = os.path.splitext(audio_path)
        audio_path = Path(fn + f"-{sample_rate}" + ext)
        torchaudio.save(audio_path, wav, sample_rate)
    if wav.size(0) != 1:
        print(f"converting audio from {wav.size(0)} channels to mono")
        wav = torch.mean(wav, dim=0).unsqueeze(0)
        fn, ext = os.path.splitext(audio_path)
        audio_path = Path(fn + f"-{sample_rate}-mono" + ext)
        torchaudio.save(audio_path, wav, sample_rate)
    print("processing text")
    sentence_list = read_text(text_path)
    transducer = create_transducer("".join(sentence_list), labels, debug)
    text_hash = TextHash(sentence_list, transducer)
    print("performing alignment")
    from .heavy import align_speech_file

    characters, words, sentences, num_frames = align_speech_file(
        wav, text_hash, model, labels, word_padding, sentence_padding
    )
    print("creating textgrid")
    waveform_to_frame_ratio = wav.size(1) / num_frames
    tg = create_text_grid_from_segments(
        characters, "characters", waveform_to_frame_ratio, sample_rate=sample_rate
    )
    words_tg = create_text_grid_from_segments(
        words, "words", waveform_to_frame_ratio, sample_rate=sample_rate
    )
    sentences_tg = create_text_grid_from_segments(
        sentences, "sentences", waveform_to_frame_ratio, sample_rate=sample_rate
    )
    tg.tiers += words_tg.get_tiers()
    tg.tiers += sentences_tg.get_tiers()
    tg_path = audio_path.with_suffix(".TextGrid")
    print(f"writing file to {tg_path}")
    tg.to_file(tg_path)


if __name__ == "__main__":
    align_single()



from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.config import (  # noqa F401
    FastSpeech2Config,
    FastSpeech2ModelConfig,
)

# If more models, change to Union[FastSpeech2Config, OtherModel]
FeaturePredictionConfig = FastSpeech2Config



import sys
from typing import Union

import simple_term_menu
from questionary import Style
from rich import print
from rich.panel import Panel

CUSTOM_QUESTIONARY_STYLE = Style(
    [
        ("qmark", "fg:default"),  # token in front of the question
        ("question", "bold"),  # question text
        ("answer", "fg:default"),  # submitted answer text behind the question
        ("pointer", "fg:default"),  # pointer used in select and checkbox prompts
        (
            "highlighted",
            "fg:default bold",
        ),  # pointed-at choice in select and checkbox prompts
        ("selected", "fg:default bold"),  # style for a selected item of a checkbox
        ("separator", "fg:default"),  # separator in lists
        ("instruction", "reverse"),  # user instructions for select, rawselect, checkbox
        ("text", "fg:default"),  # plain text
        ("disabled", "fg:default"),  # disabled choices for select and checkbox prompts
    ]
)


def get_response_from_menu_prompt(
    prompt_text: str = "",
    choices: tuple[str, ...] = (),
    title: str = "",
    multi=False,
    search=False,
    return_indices=False,
) -> Union[int, str, list[int], list[str]]:
    """Given some prompt text and a list of choices, create a simple terminal window
       and return the index of the choice

    Args:
        prompt_text (str): rich prompt text to print before menu
        choices (list[str]): choices to display

    Returns:
        int: index of choice
    """
    if prompt_text:
        print(Panel(prompt_text))
    menu = simple_term_menu.TerminalMenu(
        choices,
        title=title,
        multi_select=multi,
        multi_select_select_on_accept=(not multi),
        multi_select_empty_ok=multi,
        raise_error_on_interrupt=True,
        show_multi_select_hint=multi,
        show_search_hint=search,
        status_bar_style=("fg_gray", "bg_black"),
    )
    index = menu.show()
    sys.stdout.write("\033[K")
    if index is None or return_indices:
        return index
    else:
        if isinstance(index, tuple):
            return [choices[i] for i in index]
        else:
            return choices[index]



import doctest
import shutil
import tempfile
from pathlib import Path

import torch
import torchaudio
from pydantic_core._pydantic_core import ValidationError
from torch import float32

import everyvoice.preprocessor
import everyvoice.text.text_processor
from everyvoice.config.preprocessing_config import (
    AudioConfig,
    AudioSpecTypeEnum,
    PreprocessingConfig,
)
from everyvoice.config.shared_types import init_context
from everyvoice.model.aligner.config import AlignerConfig
from everyvoice.model.e2e.config import FeaturePredictionConfig
from everyvoice.model.vocoder.config import VocoderConfig
from everyvoice.preprocessor import Preprocessor
from everyvoice.tests.basic_test_case import BasicTestCase
from everyvoice.tests.preprocessed_audio_fixture import PreprocessedAudioFixture
from everyvoice.tests.stubs import (
    capture_stdout,
    monkeypatch,
    mute_logger,
    patch_logger,
)
from everyvoice.utils import generic_psv_filelist_reader


class PreprocessingTest(PreprocessedAudioFixture, BasicTestCase):
    """Unit tests for preprocessing steps"""

    filelist = generic_psv_filelist_reader(BasicTestCase.data_dir / "metadata.psv")

    def test_run_doctest(self):
        """Run doctests in everyvoice.text.text_processing"""
        results = doctest.testmod(everyvoice.text.text_processor)
        self.assertFalse(results.failed, results)

    # def test_compute_stats(self):
    #     feat_prediction_config = EveryVoiceConfig.load_config_from_path().feature_prediction
    #     preprocessor = Preprocessor(feat_prediction_config)
    #     preprocessor.compute_stats()
    # self.assertEqual(
    #     self.preprocessor.config["preprocessing"]["audio"]["mel_mean"],
    #     -4.018,
    #     places=3,
    # )
    # self.assertEqual(
    #     self.preprocessor.config["preprocessing"]["audio"]["mel_std"],
    #     4.017,
    #     places=3,
    # )

    def test_read_filelist(self):
        self.assertEqual(self.filelist[0]["basename"], "LJ050-0269")

    def test_no_permissions(self):
        no_permissions_args = self.fp_config.model_dump()
        no_permissions_args["preprocessing"]["source_data"][0][
            "permissions_obtained"
        ] = False
        with tempfile.TemporaryDirectory() as tmpdir:
            with init_context({"writing_config": Path(tmpdir)}):
                with self.assertRaises(ValueError):
                    FeaturePredictionConfig(**no_permissions_args)

    def test_process_audio_for_alignment(self):
        config = AlignerConfig(contact=self.contact)
        for entry in self.filelist[1:]:
            # This just applies the SOX effects
            audio, sr = self.preprocessor.process_audio(
                self.wavs_dir / (entry["basename"] + ".wav"),
                sox_effects=config.preprocessing.source_data[0].sox_effects,
                hop_size=config.preprocessing.audio.fft_hop_size,
            )
            self.assertEqual(sr, 22050)
            self.assertEqual(audio.dtype, float32)

    def test_remove_silence(self):
        audio_path_with_silence = str(
            self.data_dir / ("440tone-with-leading-trailing-silence.wav")
        )
        config = FeaturePredictionConfig(contact=self.contact)
        sox_effects = config.preprocessing.source_data[0].sox_effects + [
            [
                "silence",
                "1",
                "0.1",
                "0.1%",
            ],
            ["reverse"],  # reverse the clip to trim silence from end
            ["silence", "1", "0.1", "0.1%"],
            ["reverse"],  # reverse the clip again to revert to the right direction :)
        ]
        raw_audio, raw_sr = torchaudio.load(audio_path_with_silence)
        processed_audio, processed_sr = self.preprocessor.process_audio(
            audio_path_with_silence,
            sox_effects=sox_effects,
            hop_size=config.preprocessing.audio.fft_hop_size,
        )

        self.assertEqual(
            raw_sr, processed_sr, "Sampling Rate should not be changed by default"
        )
        self.assertEqual(
            raw_audio.size()[1] / raw_sr,
            3.5,
            "Should be exactly 3.5 seconds of audio at 44100 Hz sampling rate",
        )
        self.assertEqual(
            round(processed_audio.size()[0] / processed_sr, 2),
            2.5,
            msg="Should be about half a second of silence removed from the beginning and end",
        )
        # should work with resampling too
        rs_processed_audio, rs_processed_sr = self.preprocessor.process_audio(
            audio_path_with_silence,
            resample_rate=22050,
            sox_effects=sox_effects,
            hop_size=config.preprocessing.audio.fft_hop_size,
        )
        self.assertEqual(
            round(rs_processed_audio.size()[0] / rs_processed_sr, 2),
            2.5,
            msg="Should be about half a second of silence removed from the beginning and end when resampled too",
        )

    def test_process_empty_audio(self):
        for fn in ["empty.wav", "zeros.wav"]:
            audio, sr = self.preprocessor.process_audio(self.data_dir / fn)
            self.assertEqual(audio, None)
            self.assertEqual(sr, None)

    def test_process_audio(self):
        import torchaudio

        for entry in self.filelist[1:]:
            audio, sr = self.preprocessor.process_audio(
                self.wavs_dir / (entry["basename"] + ".wav"), hop_size=256
            )
            self.assertEqual(sr, 22050)
            self.assertEqual(audio.dtype, float32)
        # test that truncating according to hop size actually happened
        raw_audio, raw_sr = torchaudio.load(
            str(self.wavs_dir / (entry["basename"] + ".wav"))
        )
        # remove channel info
        raw_audio = raw_audio.squeeze()
        self.assertNotEqual(raw_audio.size(0), audio.size(0))
        self.assertLess(raw_audio.size(0) - audio.size(0), 256)
        # changing the hop size changes how much is removed
        diff_hop_audio, _ = self.preprocessor.process_audio(
            self.wavs_dir / (entry["basename"] + ".wav"), hop_size=35
        )
        self.assertNotEqual(audio.size(0), diff_hop_audio.size(0))
        # we should never truncate more than a portion of a single frame
        self.assertLess(raw_audio.size(0) - diff_hop_audio.size(0), 35)
        with self.assertRaises(ValueError):  # missing hop_size
            self.preprocessor.process_audio(
                self.wavs_dir / (entry["basename"] + ".wav")
            )

    def test_spectral_feats(self):
        linear_vocoder_config = VocoderConfig(
            contact=self.contact,
            preprocessing=PreprocessingConfig(
                audio=AudioConfig(spec_type=AudioSpecTypeEnum.linear)
            ),
        )
        complex_vocoder_config = VocoderConfig(
            contact=self.contact,
            preprocessing=PreprocessingConfig(
                audio=AudioConfig(spec_type=AudioSpecTypeEnum.raw)
            ),
        )
        linear_preprocessor = Preprocessor(linear_vocoder_config)
        complex_preprocessor = Preprocessor(complex_vocoder_config)

        for entry in self.filelist[1:]:
            audio, _ = self.preprocessor.process_audio(
                self.wavs_dir / (entry["basename"] + ".wav"),
                hop_size=linear_vocoder_config.preprocessing.audio.fft_hop_size,
            )

            # ming024_feats = np.load(
            #     self.data_dir
            #     / "ming024"
            #     / ("eng-LJSpeech-mel-" + entry["basename"] + ".npy")
            # )
            feats = self.preprocessor.extract_spectral_features(
                audio, self.preprocessor.input_spectral_transform
            )
            linear_feats = linear_preprocessor.extract_spectral_features(
                audio, linear_preprocessor.input_spectral_transform
            )
            complex_feats = complex_preprocessor.extract_spectral_features(
                audio, complex_preprocessor.input_spectral_transform, normalize=False
            )
            # check data is same number of mels
            self.assertEqual(
                feats.size(0),
                self.preprocessor.config.preprocessing.audio.n_mels,
            )
            # Check linear spec has right number of fft bins
            self.assertEqual(
                linear_feats.size(0),
                linear_preprocessor.config.preprocessing.audio.n_fft // 2 + 1,
            )
            # check all same length
            self.assertEqual(feats.size(1), linear_feats.size(1))
            # check all same length
            self.assertEqual(complex_feats.size(1), linear_feats.size(1))

    def test_bad_pitch(self):
        """Some files don't have any pitch values so we should make sure we handle these properly"""
        pyworld_config = FeaturePredictionConfig(
            contact=self.contact, preprocessing=PreprocessingConfig()
        )
        preprocessor_pyworld = Preprocessor(pyworld_config)
        audio = torch.zeros(22050)
        feats = self.preprocessor.extract_spectral_features(
            audio, self.preprocessor.input_spectral_transform
        )
        frame_pitch_pyworld = preprocessor_pyworld.extract_pitch(audio)
        self.assertEqual(frame_pitch_pyworld.max(), 0)
        self.assertEqual(frame_pitch_pyworld.min(), 0)
        self.assertEqual(frame_pitch_pyworld.size(0), feats.size(1))

    def test_pitch(self):
        pyworld_config = VocoderConfig(
            contact=self.contact, preprocessing=PreprocessingConfig()
        )
        preprocessor_pyworld = Preprocessor(pyworld_config)

        for entry in self.filelist[1:]:
            audio, _ = self.preprocessor.process_audio(
                self.wavs_dir / (entry["basename"] + ".wav"),
                hop_size=pyworld_config.preprocessing.audio.fft_hop_size,
            )
            dur_path = (
                self.lj_preprocessed
                / "duration"
                / self.preprocessor.sep.join(
                    [
                        entry["basename"],
                        entry.get("speaker", "default"),
                        entry.get("language", "default"),
                        "duration.pt",
                    ]
                )
            )
            durs = torch.load(dur_path)
            feats = self.preprocessor.extract_spectral_features(
                audio, self.preprocessor.input_spectral_transform
            )
            # ming024_pitch = np.load(
            #     self.data_dir
            #     / "ming024"
            #     / ("eng-LJSpeech-pitch-" + entry["basename"] + ".npy")
            # )
            # Ensure avg pitch for each phone
            frame_pitch_pyworld = preprocessor_pyworld.extract_pitch(audio)
            pyworld_phone_avg_energy = preprocessor_pyworld.average_data_by_durations(
                frame_pitch_pyworld, durs
            )
            # Ensure avg pitch for each phone
            self.assertEqual(len(durs), pyworld_phone_avg_energy.size(0))
            # Ensure same number of frames
            self.assertEqual(frame_pitch_pyworld.size(0), feats.size(1))

    # TODO: test nans: torch.any(torch.Tensor([[torch.nan, 2]]).isnan())

    def test_duration(self):
        for entry in self.filelist[1:]:
            audio, _ = self.preprocessor.process_audio(
                self.wavs_dir / (entry["basename"] + ".wav"), hop_size=256
            )
            dur_path = (
                self.lj_preprocessed
                / "duration"
                / self.preprocessor.sep.join(
                    [
                        entry["basename"],
                        entry.get("speaker", "default"),
                        entry.get("language", "default"),
                        "duration.pt",
                    ]
                )
            )
            durs = torch.load(dur_path)
            feats = self.preprocessor.extract_spectral_features(
                audio, self.preprocessor.input_spectral_transform
            )
            # ming024_durs = np.load(
            #     self.data_dir
            #     / "ming024"
            #     / ("eng-LJSpeech-duration-" + entry["basename"] + ".npy")
            # )
            # Ensure durations same number of frames as spectral features
            # note: this is off by a few frames due to mismatches in hop size between the aligner the test data
            # was trained with and the settings defined by the spectral transform function here.
            # It would be a problem if it weren't  but it's not really relevant since we're using jointly learned alignments now.
            self.assertTrue(feats.size(1) - int(sum(durs)) <= 10)

    def test_energy(self):
        frame_energy_config = VocoderConfig(
            contact=self.contact, preprocessing=PreprocessingConfig()
        )
        preprocessor = Preprocessor(frame_energy_config)
        for entry in self.filelist[1:]:
            audio, _ = self.preprocessor.process_audio(
                self.wavs_dir / (entry["basename"] + ".wav"),
                hop_size=frame_energy_config.preprocessing.audio.fft_hop_size,
            )
            dur_path = (
                self.lj_preprocessed
                / "duration"
                / self.preprocessor.sep.join(
                    [
                        entry["basename"],
                        entry.get("speaker", "default"),
                        entry.get("language", "default"),
                        "duration.pt",
                    ]
                )
            )
            durs = torch.load(dur_path)
            # ming024_energy = np.load(
            #     self.data_dir
            #     / "ming024"
            #     / ("eng-LJSpeech-energy-" + entry["basename"] + ".npy")
            # )
            feats = self.preprocessor.extract_spectral_features(
                audio, self.preprocessor.input_spectral_transform
            )

            frame_energy = preprocessor.extract_energy(feats)
            phone_avg_energy = preprocessor.average_data_by_durations(
                frame_energy, durs
            )
            # Ensure avg energy for each phone
            self.assertEqual(phone_avg_energy.size(0), len(durs))
            # Ensure same number of frames
            self.assertEqual(frame_energy.size(0), feats.size(1))

    def test_sanity(self):
        """TODO: make sanity checking code for each type of data, maybe also data analysis tooling"""
        pass

    def test_text_processing(self):
        with tempfile.TemporaryDirectory(
            prefix="test_text_processing"
        ) as tempdir, init_context({"writing_config": Path(tempdir)}):
            characters_eng_filelist = (
                self.data_dir / "metadata_characters_supported_lang.psv"
            )
            characters_default_filelist = (
                self.data_dir / "metadata_characters_no_supported_lang.psv"
            )
            arpabet_filelist = self.data_dir / "metadata_arpabet.psv"
            phones_filelist = self.data_dir / "metadata_phones.psv"
            mixed_representation_filelist = (
                self.data_dir / "metadata_mixed_representation.psv"
            )
            fp_config = FeaturePredictionConfig(**self.fp_config.model_dump())
            filelists_to_test = [
                {
                    "path": characters_eng_filelist,
                    "contains_characters": True,
                    "contains_phones": True,
                },  # will take characters and apply eng g2p
                {
                    "path": characters_default_filelist,
                    "contains_characters": True,
                    "contains_phones": False,
                },  # will just tokenize characters
                {
                    "path": arpabet_filelist,
                    "contains_characters": False,
                    "contains_phones": True,
                },  # will convert arpabet to phones
                {
                    "path": phones_filelist,
                    "contains_characters": False,
                    "contains_phones": True,
                },  # will just tokenize phones
                {
                    "path": mixed_representation_filelist,
                    "contains_characters": True,
                    "contains_phones": True,
                },  # will tokenize characters and tokenize phones
            ]
            for filelist_test_info in filelists_to_test:
                with tempfile.TemporaryDirectory(prefix="inputs", dir=".") as tmpdir:
                    tmpdir = Path(tmpdir)
                    preprocessed_dir = tmpdir / "preprocessed"
                    preprocessed_dir.mkdir(parents=True, exist_ok=True)
                    output_filelist = preprocessed_dir / "preprocessed_filelist.psv"
                    shutil.copyfile(filelist_test_info["path"], output_filelist)
                    fp_config.preprocessing.source_data[0].filelist = (
                        filelist_test_info["path"]
                    )
                    fp_config.preprocessing.save_dir = preprocessed_dir
                    preprocessor = Preprocessor(fp_config)
                    with capture_stdout() as output, mute_logger(
                        "everyvoice.preprocessor"
                    ):
                        preprocessor.preprocess(
                            output_path=str(output_filelist),
                            cpus=1,
                            to_process=["text", "pfs"],
                        )
                    self.assertIn(
                        "You've finished preprocessing: text", output.getvalue()
                    )
                    processed_filelist = preprocessor.load_filelist(output_filelist)
                    characters = [
                        x["character_tokens"]
                        for x in processed_filelist
                        if "character_tokens" in x
                    ]
                    phones = [
                        x["phone_tokens"]
                        for x in processed_filelist
                        if "phone_tokens" in x
                    ]
                    phonological_features = [
                        torch.load(f)
                        for f in sorted(
                            list((output_filelist.parent / "pfs").glob("*.pt"))
                        )
                    ]
                    for i, utt_phones in enumerate(phones):
                        # Phonlogical features are derived from phones so they should be of equal length
                        self.assertEqual(
                            len(utt_phones.split("/")),
                            phonological_features[i].size(0),
                            utt_phones.split("/"),
                        )

                    if filelist_test_info["contains_characters"]:
                        self.assertEqual(
                            len(characters),
                            5,
                            f'failed finding characters in {filelist_test_info["path"]}',
                        )
                        self.assertEqual(
                            characters[0],
                            "t/h/e/ /e/s/s/e/n/t/i/a/l/ /t/e/r/m/s/ /o/f/ /s/u/c/h/ /m/e/m/o/r/a/n/d/a/ /m/i/g/h/t/ /w/e/l/l/ /b/e/ /e/m/b/o/d/i/e/d/ /i/n/ /a/n/ /e/x/e/c/u/t/i/v/e/ /o/r/d/e/r/.",
                            f'failed in {filelist_test_info["path"]}',
                        )
                    if filelist_test_info["contains_phones"]:
                        self.assertEqual(
                            len(phones),
                            5,
                            f'failed finding phones in {filelist_test_info["path"]}',
                        )
                        if "arpabet" in filelist_test_info["path"].stem:
                            # arpabet uses space for phone boundaries
                            self.assertEqual(
                                phones[0],
                                "ð/ /ʌ/ /e/ /s/ /e/ /n/ /ʃ/ /ʌ/ /l/ /t/ /r/ /m/ /z/ /ʌ/ /v/ /s/ /ʌ/ /c/h/ /m/ /e/ /m/ /r/ /æ/ /n/ /d/ /ʌ/ /m/ /a/ɪ/ /t/ /w/ /e/ /l/ /b/ /i/ /ɪ/ /m/ /b/ /ɑ/ /d/ /i/ /d/ /ɪ/ /n/ /æ/ /n/ /ɪ/ /g/ /z/ /e/ /k/ /j/ /ʌ/ /t/ /ɪ/ /v/ /ɔ/ /r/ /d/ /r/ /.",
                            )
                        else:
                            self.assertEqual(
                                phones[0],
                                "ð/ʌ/ /ɛ/s/ɛ/n/ʃ/ʌ/l/ /t/ɜ˞/m/z/ /ʌ/v/ /s/ʌ/t/ʃ/ /m/ɛ/m/ɜ˞/æ/n/d/ʌ/ /m/a/ɪ/t/ /w/ɛ/l/ /b/i/ /ɪ/m/b/ɑ/d/i/d/ /ɪ/n/ /æ/n/ /ɪ/ɡ/z/ɛ/k/j/ʌ/t/ɪ/v/ /ɔ/ɹ/d/ɜ˞/.",
                                f'failed in {filelist_test_info["path"]}',
                            )

    def get_simple_config(self, tmpdir: str | Path):
        """Create a simple config for testing"""
        tmpdir = Path(tmpdir)
        lj_preprocessed = tmpdir / "preprocessed"
        lj_filelist = lj_preprocessed / "preprocessed_filelist.psv"

        fp_config = FeaturePredictionConfig(contact=self.contact)
        fp_config.preprocessing.source_data[0].data_dir = self.data_dir / "lj" / "wavs"

        full_filelist = self.data_dir / "metadata.psv"
        partial_filelist = tmpdir / "partial-metadata.psv"
        with open(partial_filelist, mode="w") as f_out:
            with open(full_filelist) as f_in:
                lines = list(f_in)
                for line in lines[:4]:
                    f_out.write(line)
        fp_config.preprocessing.source_data[0].filelist = full_filelist
        fp_config.preprocessing.save_dir = lj_preprocessed

        to_process = ("audio", "energy", "pitch", "attn", "text", "spec")
        return (fp_config, lj_filelist, full_filelist, partial_filelist, to_process)

    def test_incremental_preprocess(self):
        with tempfile.TemporaryDirectory(
            prefix="test_incremental_preprocess", dir="."
        ) as tmpdir:
            (
                fp_config,
                lj_filelist,
                full_filelist,
                partial_filelist,
                to_process,
            ) = self.get_simple_config(tmpdir)

            fp_config.preprocessing.source_data[0].filelist = partial_filelist
            with capture_stdout() as output, mute_logger("everyvoice.preprocessor"):
                Preprocessor(fp_config).preprocess(
                    output_path=lj_filelist, cpus=1, to_process=to_process
                )
            self.assertRegex(output.getvalue(), r"processed files *3")
            self.assertRegex(output.getvalue(), r"previously processed files *0")

            fp_config.preprocessing.source_data[0].filelist = full_filelist
            with capture_stdout() as output, mute_logger("everyvoice.preprocessor"):
                Preprocessor(fp_config).preprocess(
                    output_path=lj_filelist, cpus=1, to_process=to_process
                )
            self.assertRegex(output.getvalue(), r"processed files *2")
            self.assertRegex(output.getvalue(), r"previously processed files *3")
            with capture_stdout() as output, mute_logger("everyvoice.preprocessor"):
                Preprocessor(fp_config).preprocess(
                    output_path=lj_filelist,
                    cpus=1,
                    overwrite=True,
                    to_process=to_process,
                    debug=True,
                )
            self.assertRegex(output.getvalue(), r"processed files *5")
            self.assertRegex(output.getvalue(), r"previously processed files *0")

    def test_gotta_do_audio_first(self):
        with tempfile.TemporaryDirectory(
            prefix="test_gotta_do_audio_first", dir="."
        ) as tmpdir:
            fp_config, lj_filelist, _, _, _ = self.get_simple_config(tmpdir)

            to_process_no_audio = ("energy", "pitch", "attn", "text", "spec")
            with self.assertRaises(SystemExit), capture_stdout():
                Preprocessor(fp_config).preprocess(
                    output_path=lj_filelist, cpus=1, to_process=to_process_no_audio
                )

    def test_empty_preprocess(self):
        # Test case where the file list is not empty but after filtering
        # silence, the result is empty. The behaviour of the code base is not
        # super satisfying, we exit when we try to read
        # preprocessed/filelist.psv and it's not there, rather than catching the
        # fact that we're trying to write an empty list.
        with tempfile.TemporaryDirectory(
            prefix="test_empty_preprocess", dir="."
        ) as tmpdir:
            tmpdir = Path(tmpdir)
            fp_config, lj_filelist, _, _, to_process = self.get_simple_config(tmpdir)
            fp_config.preprocessing.source_data[0].data_dir = self.data_dir
            input_filelist = tmpdir / "empty-metadata.psv"
            with open(input_filelist, mode="w") as f:
                print("basename|raw_text|characters|speaker|language", file=f)
                print("empty|foo bar baz|foo bar baz|noone|und", file=f)
            fp_config.preprocessing.source_data[0].filelist = input_filelist

            with self.assertRaises(SystemExit), capture_stdout():
                Preprocessor(fp_config).preprocess(
                    output_path=lj_filelist, cpus=1, to_process=to_process
                )

    def test_config_lock(self):
        with tempfile.TemporaryDirectory(prefix="test_config_lock", dir=".") as tmpdir:
            tmpdir = Path(tmpdir)
            fp_config, lj_filelist, _, _, to_process = self.get_simple_config(tmpdir)

            Preprocessor(fp_config).preprocess(
                output_path=lj_filelist, cpus=1, to_process=to_process
            )

            def fail_config_lock(config_object, element, value, message):
                import everyvoice.preprocessor.preprocessor

                with monkeypatch(config_object, element, value):
                    with self.assertRaises(SystemExit):
                        with patch_logger(
                            everyvoice.preprocessor.preprocessor
                        ) as logger:
                            with self.assertLogs(logger) as logs:
                                Preprocessor(fp_config).preprocess(
                                    output_path=lj_filelist,
                                    cpus=1,
                                    to_process=to_process,
                                )
                log_output = "\n".join(logs.output)
                self.assertIn("Config lock mismatch:", log_output)
                self.assertIn(message, log_output)

            fail_config_lock(
                fp_config.preprocessing.audio,
                "alignment_sampling_rate",
                42,
                "differs from locked preprocessing.audio config",
            )

            fail_config_lock(
                fp_config.text, "cleaners", [], "differs from locked text config"
            )

            fail_config_lock(
                fp_config.preprocessing.source_data[0],
                "sox_effects",
                [],
                "differs from locked preprocessing.source_data",
            )

            Preprocessor(fp_config).save_config_lock(in_progress=True)
            fail_config_lock(
                fp_config.preprocessing,
                "_na",
                "_na",
                "the previous preprocessing run was interrupted",
            )

            lock_file = tmpdir / "preprocessed" / ".config-lock"
            lock_file.chmod(0o666)
            with open(tmpdir / "preprocessed" / ".config-lock", "w") as f:
                f.write("This is not valid JSON")
            fail_config_lock(
                fp_config.preprocessing,
                "_na",
                "_na",
                "Error loading existing config lock",
            )

    def test_train_split(self):
        """
        PreprocessingConfig's train_split should be [0., 1.].
        """
        config = PreprocessingConfig(train_split=0.5)
        self.assertEqual(config.train_split, 0.5)

        config = PreprocessingConfig(train_split=0.0)
        self.assertEqual(config.train_split, 0.0)

        config = PreprocessingConfig(train_split=1.0)
        self.assertEqual(config.train_split, 1.0)

        with self.assertRaises(ValidationError), capture_stdout() as cout:
            config = PreprocessingConfig(train_split=-0.1)
            self.assertIn("Input should be greater than or equal to 0", cout.getvalue())
        with self.assertRaises(ValidationError), capture_stdout() as cout:
            config = PreprocessingConfig(train_split=1.1)
            self.assertIn("Input should be less than or equal to 1", cout.getvalue())


class PreprocessingHierarchyTest(BasicTestCase):
    def test_hierarchy(self):
        """Unit tests for preprocessing steps"""

        with tempfile.TemporaryDirectory(prefix="test_hierarchy", dir=".") as tmpdir:
            tmpdir = Path(tmpdir)
            data_dir = Path(__file__).parent / "data"
            wavs_dir = data_dir / "hierarchy" / "wavs"
            preprocessed_dir = tmpdir / "hierarchy" / "preprocessed"
            filelist = preprocessed_dir / "preprocessed_filelist.psv"

            fp_config = FeaturePredictionConfig(contact=self.contact)
            fp_config.preprocessing.source_data[0].data_dir = wavs_dir
            fp_config.preprocessing.source_data[0].filelist = (
                data_dir / "hierarchy" / "metadata.psv"
            )
            fp_config.preprocessing.save_dir = preprocessed_dir
            preprocessor = Preprocessor(fp_config)

            with mute_logger("everyvoice.preprocessor"), capture_stdout():
                preprocessor.preprocess(
                    output_path=filelist,
                    cpus=2,
                    overwrite=True,
                    # to_process=("audio", "energy", "pitch", "text", "spec"),
                    # to_process=("audio", "text", "pfs", "spec", "attn", "energy", "pitch"),
                    to_process=("audio", "text", "spec", "attn", "energy", "pitch"),
                )
            for t in ("audio", "spec", "attn", "energy", "pitch"):
                # There are two speakers
                sources = [d.name for d in tmpdir.glob(f"**/{t}/*")]
                self.assertSetEqual(
                    set(sources), set(("LJ010", "LJ050")), f"failed for {t}"
                )
                # First speaker has one recording
                files = (
                    list(tmpdir.glob(f"**/{t}/LJ010/*.wav"))
                    if t == "audio"
                    else list(tmpdir.glob(f"**/{t}/LJ010/*.pt"))
                )
                self.assertEqual(len(files), 1)
                # Second speaker has 5 recordings
                files = (
                    list(tmpdir.glob(f"**/{t}/LJ050/*.wav"))
                    if t == "audio"
                    else list(tmpdir.glob(f"**/{t}/LJ050/*.pt"))
                )
                self.assertEqual(len(files), 5)



"""
CLI command to inspect EveryVoice's checkpoints.
"""

import json
import warnings
from collections import defaultdict
from json import JSONEncoder
from pathlib import Path
from typing import Any, Dict

import typer
from typing_extensions import Annotated

from everyvoice.base_cli.interfaces import complete_path

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    help="Extract checkpoint's hyperparameters.",
)


class CheckpointEncoder(JSONEncoder):
    """
    Helper JSON Encoder for missing `torch.Tensor` & `pydantic.BaseModel`.
    """

    def default(self, obj: Any):
        """
        Extends json to handle `torch.Tensor` and `pydantic.BaseModel`.
        """
        import torch
        from pydantic import BaseModel

        if isinstance(obj, torch.Tensor):
            return list(obj.shape)
        elif isinstance(obj, BaseModel):
            return json.loads(obj.json())
        return super().default(obj)


def summarize_statedict(ckpt: dict) -> dict:
    if "state_dict" not in ckpt:
        return {}
    model_keys: Dict[str, int] = defaultdict(int)
    for k, v in sorted(ckpt["state_dict"].items()):
        main_key = k.split(".")[0]
        model_keys[main_key] += v.numel()

    model_keys["TOTAL"] = sum(model_keys.values())
    return model_keys


def load_checkpoint(model_path: Path, minimal=True) -> Dict[str, Any]:
    """
    Loads a checkpoint and performs minor clean up of the checkpoint.
    Removes the `optimizer_states`'s `state` and `param_groups`'s `params`.
    Removes `state_dict` from the checkpoint.
    """
    import torch

    checkpoint = torch.load(str(model_path), map_location=torch.device("cpu"))

    if minimal:
        # Some clean up of useless stuff.
        if "optimizer_states" in checkpoint:
            for optimizer in checkpoint["optimizer_states"]:
                # Delete the optimizer history values.
                if "state" in optimizer:
                    del optimizer["state"]
                # These are simply values [0, len(checkpoint["optimizer_states"][0]["state"])].
                if "param_groups" in optimizer:
                    for param_group in optimizer["param_groups"]:
                        if "params" in param_group:
                            del param_group["params"]

        if "state_dict" in checkpoint:
            del checkpoint["state_dict"]

        if "callbacks" in checkpoint:
            del checkpoint["callbacks"]

        if "loops" in checkpoint:
            del checkpoint["loops"]
    return checkpoint


@app.command()
def inspect(
    model_path: Path = typer.Argument(
        ...,
        exists=True,
        dir_okay=False,
        file_okay=True,
        help="The path to your model checkpoint file.",
        autocompletion=complete_path,
    ),
    show_config: Annotated[
        bool,
        typer.Option(
            "--show-config/--no-show-config",  # noqa
            "-c/-C",  # noqa
            help="Show the configuration used during training",  # noqa
        ),
    ] = True,
    show_architecture: Annotated[
        bool,
        typer.Option(
            "--show-architecture/--no-show-architecture",  # noqa
            "-a/-A",  # noqa
            help="Show the model's architecture",  # noqa
        ),
    ] = True,
):
    """
    Given an EveryVoice checkpoint, show information about the configuration
    used during training, the model's architecture and the number of weights
    per layer and total weight count.
    """

    if show_config:
        config = json.dumps(
            load_checkpoint(model_path),
            ensure_ascii=False,
            indent=2,
            cls=CheckpointEncoder,
        )
        print(
            """
                ++++++++++++++
                    CONFIG
                ++++++++++++++
            """
        )
        print(config)

    if show_architecture:
        from torchinfo import summary

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.model import (
                    HiFiGAN,
                )

                model = HiFiGAN.load_from_checkpoint(model_path)
                print(summary(model, None, verbose=0))
            # NOTE if ANY exception is raise, that means the model couldn't be
            # loaded and we want to try another config type.  This is to "ask
            # forgiveness, not permission".
            except Exception:
                try:
                    from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.model import (
                        FastSpeech2,
                    )

                    model = FastSpeech2.load_from_checkpoint(model_path)
                    print(summary(model, None, verbose=0))
                except Exception:
                    from tabulate import tabulate

                    model = load_checkpoint(model_path, minimal=False)
                    print(
                        "We couldn't read your file, possibly because the version of EveryVoice that created it is incompatible with your installed version."
                    )
                    if sd_summary := summarize_statedict(model):
                        print(
                            f"We've tried to infer some information from your checkpoint. It appears to have {round(sd_summary['TOTAL'] / 1000000, 2)} M parameters."
                        )
                        print(
                            tabulate(
                                [[k, v] for k, v in sd_summary.items()],
                                tablefmt="rounded_grid",
                                intfmt=",",
                                headers=["LayerName", "Number of Parameters"],
                            )
                        )



import json
import tempfile
import time
from pathlib import Path
from typing import Callable
from unittest import TestCase

import yaml
from pydantic import ValidationError

from everyvoice import exceptions
from everyvoice.config.preprocessing_config import (
    AudioConfig,
    Dataset,
    PreprocessingConfig,
)
from everyvoice.config.shared_types import (
    BaseTrainingConfig,
    LoggerConfig,
    init_context,
)
from everyvoice.config.text_config import TextConfig
from everyvoice.model.aligner.config import AlignerConfig
from everyvoice.model.aligner.DeepForcedAligner.dfaligner.config import (
    DFAlignerTrainingConfig,
)
from everyvoice.model.e2e.config import E2ETrainingConfig, EveryVoiceConfig
from everyvoice.model.feature_prediction.config import FeaturePredictionConfig
from everyvoice.model.vocoder.config import VocoderConfig
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.config import (
    HiFiGANModelConfig,
    HiFiGANTrainingConfig,
)
from everyvoice.tests.basic_test_case import BasicTestCase
from everyvoice.utils import (
    expand_config_string_syntax,
    load_config_from_json_or_yaml_path,
    lower,
    nfc_normalize,
)
from everyvoice.wizard import (
    ALIGNER_CONFIG_FILENAME_PREFIX,
    PREPROCESSING_CONFIG_FILENAME_PREFIX,
    SPEC_TO_WAV_CONFIG_FILENAME_PREFIX,
    TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX,
    TEXT_TO_WAV_CONFIG_FILENAME_PREFIX,
)


def _writer_helper(model, filename):
    with open(filename, "w", encoding="utf8") as f:
        f.write(model.model_dump_json())


class ConfigTest(BasicTestCase):
    """Basic test for hyperparameter configuration"""

    def setUp(self) -> None:
        super().setUp()
        self.config = EveryVoiceConfig(
            contact=self.contact,
            aligner=AlignerConfig(contact=self.contact),
            feature_prediction=FeaturePredictionConfig(contact=self.contact),
            vocoder=VocoderConfig(contact=self.contact),
        )

    def test_from_object(self):
        """Test from object"""
        config_default = EveryVoiceConfig(
            contact=self.contact,
            aligner=AlignerConfig(contact=self.contact),
            feature_prediction=FeaturePredictionConfig(contact=self.contact),
            vocoder=VocoderConfig(contact=self.contact),
        )
        config_declared = EveryVoiceConfig(
            contact=self.contact,
            aligner=AlignerConfig(contact=self.contact),
            feature_prediction=FeaturePredictionConfig(contact=self.contact),
            vocoder=VocoderConfig(contact=self.contact),
            training=E2ETrainingConfig(),
        )
        config_32 = EveryVoiceConfig(
            contact=self.contact,
            aligner=AlignerConfig(contact=self.contact),
            feature_prediction=FeaturePredictionConfig(contact=self.contact),
            vocoder=VocoderConfig(contact=self.contact),
            training=E2ETrainingConfig(batch_size=32),
        )
        self.assertEqual(config_default.training.batch_size, 16)
        self.assertEqual(config_declared.training.batch_size, 16)
        self.assertEqual(config_32.training.batch_size, 32)

    def test_config_save_dirs(self):
        with tempfile.TemporaryDirectory(prefix="test_config_save_dirs") as tempdir:
            # Preprocessing Config
            tempdir = Path(tempdir)
            with self.assertRaises(ValidationError):
                preprocessing_config = PreprocessingConfig(save_dir=1)
            with init_context({"writing_config": tempdir}):
                preprocessing_config = PreprocessingConfig(save_dir="./bloop")
            self.assertTrue((tempdir / preprocessing_config.save_dir).exists())

    def test_config_partial(self):
        with tempfile.TemporaryDirectory(
            prefix="test_config_partial"
        ) as tempdir_str, init_context({"writing_config": Path(tempdir_str)}):
            # Preprocessing Config
            tempdir = Path(tempdir_str)
            _writer_helper(AudioConfig(), tempdir / "audio.json")
            config = PreprocessingConfig(
                path_to_audio_config_file=(tempdir / "audio.json")
            )
            self.assertTrue(isinstance(config.audio, AudioConfig))
            # bad partial
            with self.assertRaises(exceptions.InvalidConfiguration):
                with tempfile.NamedTemporaryFile(
                    prefix="test", mode="w", suffix=".yaml"
                ) as tf:
                    tf.write(" ")
                    tf.flush()
                    PreprocessingConfig(path_to_audio_config_file=tf.name)
            # Write shared:
            _writer_helper(PreprocessingConfig(), tempdir / "preprocessing.json")
            _writer_helper(TextConfig(), tempdir / "text.json")
            _writer_helper(BaseTrainingConfig(), tempdir / "training.json")
            # Aligner Config
            _writer_helper(
                AlignerConfig(contact=self.contact).training,
                tempdir / "aligner-training.json",
            )
            _writer_helper(
                AlignerConfig(contact=self.contact).model,
                tempdir / "aligner-model.json",
            )
            aligner_config = AlignerConfig(
                contact=self.contact,
                path_to_model_config_file=(tempdir / "aligner-model.json"),
                path_to_preprocessing_config_file=(tempdir / "preprocessing.json"),
                path_to_text_config_file=(tempdir / "text.json"),
                path_to_training_config_file=tempdir / "aligner-training.json",
            )
            _writer_helper(aligner_config, tempdir / "aligner.json")
            self.assertTrue(isinstance(aligner_config, AlignerConfig))
            # FP Config
            _writer_helper(
                FeaturePredictionConfig(contact=self.contact).training,
                tempdir / "fp-training.json",
            )
            _writer_helper(
                FeaturePredictionConfig(contact=self.contact).model,
                tempdir / "fp-model.json",
            )
            fp_config = FeaturePredictionConfig(
                contact=self.contact,
                path_to_model_config_file=(tempdir / "fp-model.json"),
                path_to_preprocessing_config_file=(tempdir / "preprocessing.json"),
                path_to_text_config_file=(tempdir / "text.json"),
                path_to_training_config_file=tempdir / "fp-training.json",
            )
            _writer_helper(fp_config, tempdir / "fp.json")
            self.assertTrue(isinstance(fp_config, FeaturePredictionConfig))
            # Vocoder Config
            _writer_helper(
                VocoderConfig(contact=self.contact).training,
                tempdir / "vocoder-training.json",
            )
            _writer_helper(
                VocoderConfig(contact=self.contact).model,
                tempdir / "vocoder-model.json",
            )
            vocoder_config = VocoderConfig(
                contact=self.contact,
                path_to_model_config_file=(tempdir / "vocoder-model.json"),
                path_to_preprocessing_config_file=(tempdir / "preprocessing.json"),
                path_to_training_config_file=tempdir / "vocoder-training.json",
            )
            _writer_helper(vocoder_config, tempdir / "vocoder.json")
            self.assertTrue(isinstance(vocoder_config, VocoderConfig))
            # E2E Config
            e2e_config = EveryVoiceConfig(
                contact=self.contact,
                path_to_aligner_config_file=(tempdir / "aligner.json"),
                path_to_feature_prediction_config_file=(tempdir / "fp.json"),
                path_to_training_config_file=(tempdir / "training.json"),
                path_to_vocoder_config_file=(tempdir / "vocoder.json"),
            )
            self.assertTrue(isinstance(e2e_config, EveryVoiceConfig))

    def test_config_partial_override(self):
        """Test override of partial"""
        with tempfile.NamedTemporaryFile(
            prefix="test_config_partial_override", mode="w", suffix=".yaml"
        ) as tf:
            tf.write(AudioConfig().model_dump_json())
            tf.flush()
            # override with actual class
            config = PreprocessingConfig(
                path_to_audio_config_file=tf.name,
                audio=AudioConfig(min_audio_length=1.0),
            )
            self.assertEqual(config.audio.min_audio_length, 1.0)
            # override with dict
            config = PreprocessingConfig(
                path_to_audio_config_file=tf.name, audio={"max_audio_length": 1.0}
            )
            self.assertEqual(config.audio.max_audio_length, 1.0)
            # pass something invalid
            with self.assertRaises(ValidationError):
                PreprocessingConfig(path_to_audio_config_file=tf.name, audio=1.0)

    def test_update_from_file(self):
        """Test that updating the config from yaml/json works"""
        with open(self.data_dir / "update.json") as f:
            update = json.load(f)
        self.config.update_config(update)
        with open(self.data_dir / "update.yaml") as f:
            update = yaml.safe_load(f)
        self.config.update_config(update)
        self.assertEqual(self.config.feature_prediction.training.batch_size, 123)
        self.assertEqual(self.config.vocoder.training.batch_size, 456)

    def test_string_to_callable(self):
        # Test Basic Functionality
        config = FeaturePredictionConfig(
            contact=self.contact, text=TextConfig(cleaners=["everyvoice.utils.lower"])
        )
        self.assertEqual(config.text.cleaners, [lower])
        # Test missing function
        with self.assertRaises(AttributeError):
            config.update_config({"text": {"cleaners": ["everyvoice.utils.foobarfoo"]}})
        # Test missing module
        with self.assertRaises(ImportError):
            config.update_config({"text": {"cleaners": ["foobarfoo.utils.lower"]}})
        # Test not string
        with self.assertRaises(ValueError):
            config.update_config({"text": {"cleaners": [1]}})
        # Test plain string
        config = LoggerConfig(sub_dir_callable="foobar")
        self.assertEqual(config.sub_dir_callable(), "foobar")

    def test_call_sub_dir(self):
        config = LoggerConfig()
        # sub_dir should get called from sub_dir_callable and be a string of an int
        self.assertTrue(isinstance(int(config.sub_dir), int))
        # Just in case we're super speedy
        time.sleep(1)
        self.assertGreater(int(config.sub_dir_callable()), int(config.sub_dir))
        serialized_config = config.model_dump()
        # Exclude sub_dir by default when serializing as it should get overriden on each run
        self.assertTrue("sub_dir" not in serialized_config)

    def test_properly_deserialized_callables(self):
        config = TextConfig(cleaners=[nfc_normalize, "everyvoice.utils.lower"])
        for fn in config.cleaners:
            self.assertTrue(isinstance(fn, Callable))

    def test_string_to_dict(self):
        base_config = EveryVoiceConfig(
            contact=self.contact,
            aligner=AlignerConfig(contact=self.contact),
            feature_prediction=FeaturePredictionConfig(contact=self.contact),
            vocoder=VocoderConfig(contact=self.contact),
        )
        test_string = "vocoder.training.gan_type=wgan"
        test_bad_strings = [
            "vocoder.training.gan_type==wgan",
            "vocoder.training.gan_typewgan",
        ]
        # test_missing = ["training.foobar.gan_type=original"]
        test_dict = expand_config_string_syntax(test_string)
        self.assertEqual(test_dict, {"vocoder": {"training": {"gan_type": "wgan"}}})
        for bs in test_bad_strings:
            with self.assertRaises(ValueError):
                expand_config_string_syntax(bs)

        self.assertEqual(base_config.vocoder.training.gan_type.value, "original")
        config = base_config.combine_configs(base_config, test_dict)
        self.assertEqual(config["vocoder"]["training"]["gan_type"], "wgan")

    def test_changes(self):
        """Test that the changes to the config are correct"""
        self.config.update_config(
            {"feature_prediction": {"text": {"cleaners": ["everyvoice.utils.lower"]}}}
        )
        self.assertEqual(self.config.feature_prediction.text.cleaners, [lower])

    def test_load_empty_config(self):
        with tempfile.NamedTemporaryFile(
            prefix="test_load_empty_config", mode="w", suffix=".yaml"
        ) as tf:
            tf.write(" ")
            tf.flush()
            with self.assertRaises(exceptions.InvalidConfiguration):
                load_config_from_json_or_yaml_path(Path(tf.name))

    def test_change_with_indices(self):
        """Text the --config-args can also work with arrays"""
        with tempfile.TemporaryDirectory(
            prefix="test_change_with_indices"
        ) as tempdir, init_context({"writing_config": Path(tempdir)}):
            config = FeaturePredictionConfig(contact=self.contact)
            config.update_config(
                {
                    "preprocessing": {
                        "source_data": {"0": {"filelist": "/foo/bar/filelist.psv"}}
                    }
                }
            )
            self.assertEqual(
                str(config.preprocessing.source_data[0].filelist),
                "/foo/bar/filelist.psv",
            )

    def test_shared_sox(self) -> None:
        """Test that the shared sox config is correct"""
        vocoder_config = VocoderConfig(
            contact=self.contact,
            preprocessing=PreprocessingConfig(
                source_data=[
                    Dataset(permissions_obtained=True),
                    Dataset(permissions_obtained=True),
                    Dataset(permissions_obtained=True),
                    Dataset(permissions_obtained=True),
                ]
            ),
        )
        config: EveryVoiceConfig = EveryVoiceConfig(
            vocoder=vocoder_config,
            contact=self.contact,
            aligner=AlignerConfig(contact=self.contact),
            feature_prediction=FeaturePredictionConfig(contact=self.contact),
        )
        sox_effects = config.vocoder.preprocessing.source_data[0].sox_effects
        self.assertEqual(len(config.vocoder.preprocessing.source_data), 4)
        for d_other in config.vocoder.preprocessing.source_data[1:]:
            self.assertEqual(sox_effects, d_other.sox_effects)
            self.assertEqual(sox_effects[0], ["channels", "1"])

    def test_correct_number_typing(self):
        batch_size = 64.0
        config = EveryVoiceConfig(
            training=E2ETrainingConfig(batch_size=batch_size),
            contact=self.contact,
            aligner=AlignerConfig(contact=self.contact),
            feature_prediction=FeaturePredictionConfig(contact=self.contact),
            vocoder=VocoderConfig(contact=self.contact),
        )
        self.assertIsInstance(batch_size, float)
        self.assertEqual(config.training.batch_size, 64)
        self.assertIsInstance(config.feature_prediction.training.batch_size, int)


class LoadConfigTest(BasicTestCase):
    """Load configs that contains relative paths."""

    REL_DATA_DIR = Path(__file__).parent / "data" / "relative" / "config"
    DATASET_NAME: str = "relative"

    def validate_config_path(self, path: Path):
        """
        Helper method to validate a path once loaded by a config.
        """
        self.assertTrue(path.is_absolute(), msg=path)
        self.assertTrue(path.exists(), msg=path)

    def test_aligner_config(self):
        """Create a AlignerConfig which pydantic will validate for us."""
        config_path = self.REL_DATA_DIR / f"{ALIGNER_CONFIG_FILENAME_PREFIX}.yaml"
        with config_path.open("r", encoding="utf8") as f:
            pre_test = yaml.safe_load(f)
            self.assertFalse(
                Path(pre_test["path_to_preprocessing_config_file"]).is_absolute()
            )
            self.assertFalse(Path(pre_test["path_to_text_config_file"]).is_absolute())
            training = pre_test["training"]
            self.assertFalse(Path(training["logger"]["save_dir"]).is_absolute())
            self.assertFalse(Path(training["training_filelist"]).is_absolute())
            self.assertFalse(Path(training["validation_filelist"]).is_absolute())
        config = AlignerConfig.load_config_from_path(config_path)
        # print(config.model_dump_json(indent=2))
        self.assertTrue(isinstance(config, AlignerConfig))
        self.assertEqual(config.preprocessing.dataset, self.DATASET_NAME)
        self.validate_config_path(config.path_to_preprocessing_config_file)
        self.validate_config_path(config.path_to_text_config_file)
        self.validate_config_path(config.training.logger.save_dir)
        self.validate_config_path(config.training.training_filelist)
        self.validate_config_path(config.training.validation_filelist)

    def test_preprocessing_config(self):
        """Create a PreprocessingConfig which pydantic will validate for us."""
        config_path = self.REL_DATA_DIR / f"{PREPROCESSING_CONFIG_FILENAME_PREFIX}.yaml"
        with config_path.open("r", encoding="utf8") as f:
            pre_test = yaml.safe_load(f)
            self.assertFalse(Path(pre_test["save_dir"]).is_absolute())
            self.assertEqual(len(pre_test["source_data"]), 1)
            for data in pre_test["source_data"]:
                self.assertFalse(Path(data["data_dir"]).is_absolute())
                self.assertFalse(Path(data["filelist"]).is_absolute())
        config = PreprocessingConfig.load_config_from_path(config_path)
        # print(config.model_dump_json(indent=2))
        self.assertTrue(isinstance(config, PreprocessingConfig))
        self.assertEqual(config.dataset, self.DATASET_NAME)
        self.validate_config_path(config.save_dir)
        self.assertEqual(len(config.source_data), 1)
        for data in config.source_data:
            self.validate_config_path(data.data_dir)
            self.validate_config_path(data.filelist)

    def test_feature_prediction_config(self):
        """Create a FeaturePredictionConfig which pydantic will validate for us."""
        config_path = self.REL_DATA_DIR / f"{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}.yaml"
        with config_path.open("r", encoding="utf8") as f:
            pre_test = yaml.safe_load(f)
            self.assertFalse(
                Path(pre_test["path_to_preprocessing_config_file"]).is_absolute()
            )
            self.assertFalse(Path(pre_test["path_to_text_config_file"]).is_absolute())
            training = pre_test["training"]
            self.assertFalse(Path(training["logger"]["save_dir"]).is_absolute())
            self.assertFalse(Path(training["training_filelist"]).is_absolute())
            self.assertFalse(Path(training["validation_filelist"]).is_absolute())
        config = FeaturePredictionConfig.load_config_from_path(config_path)
        # print(config.model_dump_json(indent=2))
        self.assertEqual(config.preprocessing.dataset, self.DATASET_NAME)
        self.validate_config_path(config.path_to_text_config_file)
        self.validate_config_path(config.path_to_text_config_file)
        self.validate_config_path(config.training.logger.save_dir)
        self.validate_config_path(config.training.training_filelist)
        self.validate_config_path(config.training.validation_filelist)

    def test_vocoder_config(self):
        """Create a VocoderConfig which pydantic will validate for us."""
        config_path = self.REL_DATA_DIR / f"{SPEC_TO_WAV_CONFIG_FILENAME_PREFIX}.yaml"
        with config_path.open("r", encoding="utf8") as f:
            pre_test = yaml.safe_load(f)
            self.assertFalse(
                Path(pre_test["path_to_preprocessing_config_file"]).is_absolute()
            )
            training = pre_test["training"]
            self.assertFalse(Path(training["logger"]["save_dir"]).is_absolute())
            self.assertFalse(Path(training["training_filelist"]).is_absolute())
            self.assertFalse(Path(training["validation_filelist"]).is_absolute())
        config = VocoderConfig.load_config_from_path(config_path)
        # print(config.model_dump_json(indent=2))
        self.assertTrue(isinstance(config, VocoderConfig))
        self.assertEqual(config.preprocessing.dataset, self.DATASET_NAME)
        self.validate_config_path(config.path_to_preprocessing_config_file)
        self.validate_config_path(config.training.logger.save_dir)
        self.validate_config_path(config.training.training_filelist)
        self.validate_config_path(config.training.validation_filelist)

    def test_everyvoice_config(self):
        """Create a EveryVoiceConfig which pydantic will validate for us."""
        config_path = self.REL_DATA_DIR / f"{TEXT_TO_WAV_CONFIG_FILENAME_PREFIX}.yaml"
        with config_path.open("r", encoding="utf8") as f:
            pre_test = yaml.safe_load(f)
            self.assertFalse(
                Path(pre_test["path_to_aligner_config_file"]).is_absolute()
            )
            self.assertFalse(
                Path(pre_test["path_to_feature_prediction_config_file"]).is_absolute()
            )
            self.assertFalse(
                Path(pre_test["path_to_vocoder_config_file"]).is_absolute()
            )
            training = pre_test["training"]
            self.assertFalse(Path(training["logger"]["save_dir"]).is_absolute())
            self.assertFalse(Path(training["training_filelist"]).is_absolute())
            self.assertFalse(Path(training["validation_filelist"]).is_absolute())
        config = EveryVoiceConfig.load_config_from_path(config_path)
        # print(config.model_dump_json(indent=2))
        self.assertTrue(isinstance(config, EveryVoiceConfig))
        self.assertEqual(
            config.feature_prediction.preprocessing.dataset, self.DATASET_NAME
        )
        self.validate_config_path(config.path_to_aligner_config_file)
        self.validate_config_path(config.path_to_feature_prediction_config_file)
        self.validate_config_path(config.path_to_vocoder_config_file)
        self.validate_config_path(config.training.logger.save_dir)
        self.validate_config_path(config.training.training_filelist)
        self.validate_config_path(config.training.validation_filelist)

    def test_absolute_path(self):
        """Load a config that has absolute paths."""
        with tempfile.TemporaryDirectory(
            prefix="test_absolute_path"
        ) as tempdir, init_context({"writing_config": Path(tempdir)}):
            tempdir = Path(tempdir).absolute()
            # Write preprocessing:
            preprocessing_config_path = tempdir / "aligner-preprocessing.json"
            _writer_helper(
                PreprocessingConfig(dataset=self.DATASET_NAME),
                preprocessing_config_path,
            )

            # Write text:
            text_config_path = tempdir / "aligner-text.json"
            _writer_helper(TextConfig(), text_config_path)

            # Write training:
            aligner_training_path = tempdir / "aligner-training.json"
            training = DFAlignerTrainingConfig(
                training_filelist=tempdir / "training_filelist.psv",
                validation_filelist=tempdir / "validation_filelist.psv",
            )
            (tempdir / training.logger.save_dir).mkdir(parents=True, exist_ok=True)
            print(tempdir, training.training_filelist)
            (training.training_filelist).touch(exist_ok=True)
            (training.validation_filelist).touch(exist_ok=True)
            _writer_helper(training, aligner_training_path)

            # Write model:
            aligner_model_path = tempdir / "aligner-model.json"
            _writer_helper(
                AlignerConfig(contact=self.contact).model, aligner_model_path
            )

            # Aligner Config
            aligner_config = AlignerConfig(
                contact=self.contact,
                path_to_model_config_file=aligner_model_path,
                path_to_preprocessing_config_file=preprocessing_config_path,
                path_to_text_config_file=text_config_path,
                path_to_training_config_file=aligner_training_path,
            )
            self.assertTrue(isinstance(aligner_config, AlignerConfig))
            aligner_config_path = tempdir / "aligner.json"
            _writer_helper(aligner_config, aligner_config_path)

            # Reload and validate
            config = AlignerConfig.load_config_from_path(aligner_config_path)
            self.assertTrue(isinstance(config, AlignerConfig))
            self.assertEqual(config.preprocessing.dataset, self.DATASET_NAME)
            self.validate_config_path(config.path_to_model_config_file)
            self.validate_config_path(config.path_to_preprocessing_config_file)
            self.validate_config_path(config.path_to_text_config_file)
            self.validate_config_path(config.path_to_training_config_file)
            self.validate_config_path(config.training.logger.save_dir)
            self.validate_config_path(config.training.training_filelist)
            self.validate_config_path(config.training.validation_filelist)

    def test_missing_path(self):
        """Load a config that is missing a partial config file."""
        with tempfile.TemporaryDirectory(
            prefix="test_missing_path"
        ) as tempdir, init_context({"writing_config": Path(tempdir)}):
            tempdir = Path(tempdir)
            _writer_helper(AudioConfig(), tempdir / "audio.json")
            config = PreprocessingConfig(
                path_to_audio_config_file=(tempdir / "audio.json")
            )
            self.assertTrue(isinstance(config.audio, AudioConfig))
            # Write shared:
            _writer_helper(
                PreprocessingConfig(dataset=self.DATASET_NAME),
                tempdir / "preprocessing.json",
            )
            _writer_helper(TextConfig(), tempdir / "text.json")
            _writer_helper(BaseTrainingConfig(), tempdir / "training.json")
            # Aligner Config
            _writer_helper(
                AlignerConfig(contact=self.contact).training,
                tempdir / "aligner-training.json",
            )
            _writer_helper(
                AlignerConfig(contact=self.contact).model,
                tempdir / "aligner-model.json",
            )
            aligner_config = AlignerConfig(
                contact=self.contact,
                path_to_model_config_file=tempdir / "aligner-model.json",
                path_to_preprocessing_config_file=tempdir / "preprocessing.json",
                path_to_text_config_file=tempdir / "text.json",
                path_to_training_config_file=tempdir / "aligner-training.json",
            )
            _writer_helper(aligner_config, tempdir / "aligner.json")
            self.assertTrue(isinstance(aligner_config, AlignerConfig))
            # Create the missing partial config file by deleting.
            # NOTE, we need the file to exists if we want to write its parent config to disk.
            (tempdir / "preprocessing.json").unlink()
            with self.assertRaises(ValidationError):
                config = AlignerConfig.load_config_from_path(tempdir / "aligner.json")


class BaseTrainingConfigTest(TestCase):
    """
    Validate BaseTrainingConfig
    """

    def test_ckpt_epochs_cannot_be_negative(self):
        """
        every_n_epochs aka ckpt_epochs must be None or non-negative.
        """
        config = BaseTrainingConfig(ckpt_epochs=None, ckpt_steps=None)
        self.assertEqual(config.ckpt_epochs, None)

        config = BaseTrainingConfig(ckpt_epochs=0, ckpt_steps=None)
        self.assertEqual(config.ckpt_epochs, 0)

        config = BaseTrainingConfig(ckpt_epochs=10, ckpt_steps=None)
        self.assertEqual(config.ckpt_epochs, 10)

        with self.assertRaises(ValueError):
            _ = BaseTrainingConfig(ckpt_epochs=-1, ckpt_steps=None)

    def test_ckpt_steps_cannot_be_negative(self):
        """
        every_n_train_steps aka ckpt_steps must be None or non-negative.
        """
        config = BaseTrainingConfig(ckpt_epochs=None, ckpt_steps=None)
        self.assertEqual(config.ckpt_steps, None)

        config = BaseTrainingConfig(ckpt_epochs=None, ckpt_steps=0)
        self.assertEqual(config.ckpt_steps, 0)

        config = BaseTrainingConfig(ckpt_epochs=None, ckpt_steps=10)
        self.assertEqual(config.ckpt_steps, 10)

        with self.assertRaises(ValueError):
            _ = BaseTrainingConfig(ckpt_epochs=None, ckpt_steps=-1)

    def test_mutually_exclusive_ckpt_options(self):
        """
        ckpt_epochs and ckpt_steps must be mutually exclusive.
        """
        config = BaseTrainingConfig()
        self.assertEqual(config.ckpt_epochs, 1)
        self.assertEqual(config.ckpt_steps, None)

        config = BaseTrainingConfig(ckpt_epochs=None, ckpt_steps=None)
        self.assertEqual(config.ckpt_epochs, None)
        self.assertEqual(config.ckpt_steps, None)

        config = BaseTrainingConfig(ckpt_epochs=7, ckpt_steps=None)
        self.assertEqual(config.ckpt_epochs, 7)
        self.assertEqual(config.ckpt_steps, None)

        config = BaseTrainingConfig(ckpt_epochs=None, ckpt_steps=11)
        self.assertEqual(config.ckpt_epochs, None)
        self.assertEqual(config.ckpt_steps, 11)

        with self.assertRaises(ValueError):
            _ = BaseTrainingConfig(
                ckpt_epochs=1,
                ckpt_steps=1,
            )


class HiFiGANModelConfigTest(TestCase):
    """
    This class should really be under HiFiGan.
    """

    def test_invalide_resblock(self):
        """
        Validate that we get a nice error message when the user provides a
        resblock that is not part of the enum's values.
        """
        from pydantic_core._pydantic_core import ValidationError

        with self.assertRaisesRegex(
            ValidationError,
            r"Input should be '1' or '2' \[type=enum, input_value='BAD', input_type=str\]",
        ):
            HiFiGANModelConfig(resblock="BAD")


class HiFiGANTrainingConfigTest(TestCase):
    """
    This class should really be under HiFiGan.
    """

    def test_invalid_gan_type(self):
        """
        Validate that we get a nice error message when the user provides a
        gan_type that is not part of the enum's values.
        """
        from pydantic_core._pydantic_core import ValidationError

        with self.assertRaisesRegex(
            ValidationError,
            r"Input should be 'original' or 'wgan' \[type=enum, input_value='BAD', input_type=str\]",
        ):
            HiFiGANTrainingConfig(gan_type="BAD")



import re

import grapheme
from ipatok import tokenise

from everyvoice.config.utils import PossiblySerializedCallable
from everyvoice.exceptions import ConfigError


def normalize_text_helper(
    text: str,
    to_replace: dict[str, str],
    cleaners: list[PossiblySerializedCallable],
    apply_replace_rules: bool = True,
    apply_cleaners: bool = True,
) -> str:
    """Helper for TextProcessor.normalize_text. We need to use it for the Symbols validator

    Args:
        text (str): The text to be normalized
        to_replace (dict[str, str]): replace rules to apply
        cleaners (list[PossiblySerializedCallable]): Cleaner functions to apply
        apply_replace_rules (bool): Whether to apply replace rules
        apply_cleaners (bool): Whether to apply cleaners

    Returns:
        str: the normalized text
    """
    if apply_replace_rules:
        text = apply_to_replace_helper(text, to_replace)
    if apply_cleaners:
        text = apply_cleaners_helper(text, cleaners)
    return text


def apply_to_replace_helper(text: str, to_replace: dict[str, str]) -> str:
    """Helper for TextProcessor.apply_replacement_rules. We need to use it for the Symbols validator

    Args:
        text (str): The text to be converted
        to_replace (dict[str, str]): replace rules to apply

    Returns:
        str: The text with replace rules applied
    """
    for k, v in to_replace.items():
        text = re.sub(k, v, text)
    return text


def apply_cleaners_helper(text: str, cleaners: list[PossiblySerializedCallable]) -> str:
    """Helper for TextProcesser.apply_cleaners. We need to use it for the Symbols validator

    Args:
        text (str): The text to be converted
        cleaners (list[PossiblySerializedCallable]): Cleaner functions to apply

    Raises:
        ConfigError: a cleaner threw an error

    Returns:
        str: cleaned text
    """
    for cleaner_fn in cleaners:
        try:
            text = cleaner_fn(text)
        except Exception as e:
            raise ConfigError(f"Cleaner did not work and threw exception {e}") from e
    return text


def guess_graphemes_in_text(text: str) -> set[str]:
    """Given some text, determine the set of graphemes by
        applying Unicode grapheme clustering rules.

    Args:
        text (str): some normalized, un-tokenized text
    Returns:
        set[str]: a set of possible graphemes

    >>> sorted(guess_graphemes_in_text('g\u0331an'))
    ['a', 'g̱', 'n']
    """
    return set(grapheme.graphemes(text))


def guess_graphemes_in_text_lines(text_lines: list[str]) -> set[str]:
    """Helper function for guessing graphemes in multiple lines of text,
        determined by applying Unicode grapheme clustering rules.

    Args:
        text (list[str]): some normalized, un-tokenized text separated line-by-line
    Returns:
        set[str]: a set of possible graphemes

    >>> example_data = [" කෝකටත් මං වෙනදා ", " ඇන්ජලීනා ජොලී කියන්නේ "]
    >>> sorted(guess_graphemes_in_text_lines(example_data))
    [' ', 'ඇ', 'ක', 'කි', 'කෝ', 'ජ', 'ජො', 'ට', 'ත්', 'දා', 'න', 'න්', 'නා', 'නේ', 'මං', 'ය', 'ලී', 'වෙ']
    """
    graphemes = set()
    for line in text_lines:
        graphemes.update(guess_graphemes_in_text(line))
    return graphemes


def guess_ipa_phones_in_text(text: str) -> set[str]:
    """Given some text, determine the set of valid ipa phones
        by applying strict IPA tokenization and discarding the rest

    Args:
        text (str): some normalized, un-tokenized text that has IPA characters
    Returns:
        set[str]: a set of possible IPA phones

    >>> sorted(guess_ipa_phones_in_text('ʃin1[}!]'))
    ['i', 'n', 'ʃ']

    """
    return set(tokenise(text, replace=False, tones=True, strict=False, unknown=False))


def guess_ipa_phones_in_text_lines(text_lines: list[str]) -> set[str]:
    """Given some text, determine the set of valid ipa phones
        by applying strict IPA tokenization and discarding the rest

    Args:
        text (list[str]): some normalized, un-tokenized text that has IPA characters separated line-by-line
    Returns:
        set[str]: a set of possible IPA phones

    # TODO: panphon doesn't agree with ipatok about whether g is valid IPA

    >>> example_data = ["ʃin", "gotcha"]
    >>> sorted(guess_graphemes_in_text_lines(example_data))
    ['a', 'c', 'g', 'h', 'i', 'n', 'o', 't', 'ʃ']
    """
    phones = set()
    for line in text_lines:
        phones.update(guess_ipa_phones_in_text(line))
    return phones



#!/usr/bin/env python

""" Organize tests into Test Suites
"""

import importlib
import os
import re
import sys
from typing import Iterable
from unittest import TestLoader, TestSuite, TextTestRunner

from loguru import logger

try:
    # Figure out if we are running in an environment where we've installed  everyvoice or fs2?
    import fs2  # noqa: F401

    PREFIX = "fs2."
except ImportError:
    PREFIX = "everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2."

# Unit tests

SUITES: dict[str, tuple[str, ...]] = {
    "cli": ("test_cli",),
    "callbacks": ("test_writing_callbacks",),
}
dev_suites = (
    "cli",
    "callbacks",
)
SUITES["dev"] = sum((SUITES[suite] for suite in dev_suites), start=())


def remove_test_prefix(test_case: str):
    for prefix in "<", PREFIX, "tests.":
        if test_case.startswith(prefix):
            test_case = test_case[len(prefix) :]
    return "<" + test_case


def list_tests(suite: TestSuite):
    for subsuite in suite:
        # print(str(subsuite))
        for match in re.finditer(r"tests=\[([^][]+)\]>", str(subsuite)):
            for test_case in match[1].split(", "):
                yield remove_test_prefix(test_case)


def all_test_suites() -> TestSuite:
    loader = TestLoader()
    # NOTE: Looking specifically under `/tests` removes empty TestSuites.
    test_suite = loader.discover(
        os.path.dirname(__file__) + "/tests",
        top_level_dir=os.path.dirname(__file__),
    )

    return test_suite


def describe_suite(suite: TestSuite):
    full_suite = all_test_suites()
    full_list = list(list_tests(full_suite))
    requested_list = list(list_tests(suite))
    requested_set = set(requested_list)
    print("Test suite includes:", *sorted(requested_list), sep="\n")
    print(
        "\nTest suite excludes:",
        *sorted(set(full_list) - requested_set),
        sep="\n",
    )


def run_tests(suite: str, describe: bool = False):
    """Decide which Test Suite to run"""
    logger.info(f"Loading test suite '{suite}'.")
    if suite == "all":
        test_suite = all_test_suites()
    else:
        loader = TestLoader()
        tests: Iterable[str]
        if suite in SUITES:
            tests = SUITES[suite]
        else:
            logger.error(
                f"Please specify a test suite to run: one of '{['all'] + sorted(SUITES.keys())}'."
            )
            return False
        tests = [
            PREFIX + "tests." + test if not test.startswith("/") else test
            for test in tests
        ]
        test_suite = TestSuite()
        for test in tests:
            if test.startswith("/"):
                sub_suite = loader.discover(
                    os.path.dirname(__file__) + test,
                    top_level_dir=os.path.dirname(__file__),  # MANDATORY
                )
                test_suite.addTests(sub_suite)
            else:
                importlib.import_module(test)
                test_suite.addTest(loader.loadTestsFromName(test))

    if describe:
        describe_suite(test_suite)
        return True
    else:
        logger.info("Running test suite")
        return TextTestRunner(verbosity=3).run(test_suite).wasSuccessful()


if __name__ == "__main__":
    describe = "--describe" in sys.argv
    if describe:
        sys.argv.remove("--describe")

    try:
        suite = sys.argv[1]
    except IndexError:
        logger.info('No test suite specified, defaulting to "dev"')
        suite = "dev"
    result = run_tests(suite, describe)
    if not result:
        sys.exit(1)



import json
from enum import Enum
from pathlib import Path
from typing import Any

import typer

from everyvoice._version import VERSION
from everyvoice.base_cli.checkpoint import inspect as inspect_checkpoint
from everyvoice.base_cli.interfaces import complete_path
from everyvoice.model.aligner.wav2vec2aligner.aligner.cli import (
    align_single as ctc_segment,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.cli.preprocess import (
    preprocess as preprocess_fs2,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.cli.synthesize import (
    synthesize as synthesize_fs2,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.cli.train import (
    train as train_fs2,
)
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.cli import (
    synthesize as synthesize_hfg,
)
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.cli import train as train_hfg
from everyvoice.wizard import (
    ALIGNER_CONFIG_FILENAME_PREFIX,
    PREPROCESSING_CONFIG_FILENAME_PREFIX,
    SPEC_TO_WAV_CONFIG_FILENAME_PREFIX,
    TEXT_CONFIG_FILENAME_PREFIX,
    TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX,
    TEXT_TO_WAV_CONFIG_FILENAME_PREFIX,
)


# See https://github.com/tiangolo/typer/issues/428#issuecomment-1238866548
class TyperGroupOrderAsDeclared(typer.core.TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()


app = typer.Typer(
    pretty_exceptions_show_locals=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    rich_markup_mode="markdown",
    cls=TyperGroupOrderAsDeclared,
    help="""
    # Welcome to the EveryVoice Command Line Interface

    Please visit the documentation to get started: [https://docs.everyvoice.ca](https://docs.everyvoice.ca)

    ## Use your own language/dataset

    To run the wizard for a new project please use the following command: everyvoice new-project

    ## Segment long files in your dataset

    If you have long audio files that contain more than one utterance,
    you can use the segmentation tool by running everyvoice segment [OPTIONS]

    ## Preprocess

    Once you have a configuration, preprocess your data by running everyvoice preprocess [OPTIONS]

    ## Train

    Once you have a configuration and have preprocessed your data, train a model by running everyvoice train [text-to-spec|spec-to-wav] [OPTIONS].
    EveryVoice has different types of models you can train:

    1. **text-to-spec** --- this is the most common model you will need to train. It is a model from text inputs to spectral feature (aka spectrogram) outputs.

    2. **spec-to-wav** --- this is the model that turns your spectral features into audio. It is also known as a 'vocoder'. You will typically not need to train your own version. Please refer to [https://pathtocheckpoints](https://pathtocheckpoints) for more information.

    ## Synthesize

    Once you have a trained model, generate some audio by running: everyvoice synthesize [text-to-spec|spec-to-wav] [OPTIONS]
""",
)


class ModelTypes(str, Enum):
    text_to_spec = "text-to-spec"
    spec_to_wav = "spec-to-wav"


app.command(
    short_help="Segment a long audio file",
    name="segment",
    help="""
    # Segmentation help

    This command will segment a long audio file into multiple utterances which is required for training a TTS system.
    This command should work on most languages and you should run it before running the new project or preprocessing steps.
    """,
)(ctc_segment)


@app.command(
    short_help="This command will help you create all the configuration necessary for using a new project.",
    help="""
    # This command will help you create all the configuration necessary for using a new project.

    ## Getting Started

    In order to get started, please:

        1. make sure your audio data is available on your computer and in .wav format

        2. have a 'metadata' file that minimally has two columns, one for the text of the audio and one for the basename of the file


    ## Extra details and examples

    Inside /path/to/wavs/ you should have wav audio files like test0001.wav, test0002.wav etc - they can be called anything you want but the part of the file (minus the .wav portion) must be in your metadata file.

    Example wavs directory: "/path/to/wavs/"

    Example metadata: [https://github.com/roedoejet/EveryVoice/blob/main/everyvoice/filelists/lj_full.psv](https://github.com/roedoejet/EveryVoice/blob/main/everyvoice/filelists/lj_full.psv)

""",
)
def new_project():
    from everyvoice.wizard.main_tour import get_main_wizard_tour

    get_main_wizard_tour().run()


# Add preprocess to root
app.command(
    short_help="Preprocess your data",
    help=f"""
    # Preprocess Help

    This command will preprocess all of the data you need for use with EveryVoice.

    By default every step of the preprocessor will be done by running:
    \n\n
    **everyvoice preprocess config/{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}.yaml**
    \n\n
    If you only want to process specific things, you can run specific commands by adding them as options for example:
    \n\n
    **everyvoice preprocess config/{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}.yaml -s energy -s pitch**
    """,
)(preprocess_fs2)

# Add the train commands
train_group = typer.Typer(
    pretty_exceptions_show_locals=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    rich_markup_mode="markdown",
    cls=TyperGroupOrderAsDeclared,
    help="""
    # Train Help

        - **text-to-spec** --- this is the most common model you will need to train. It is a model from text inputs to spectral feature (aka spectrogram) outputs.

        - **spec-to-wav** --- this is the model that turns your spectral features into audio. It is also known as a 'vocoder'. You will typically not need to train your own version. Please refer to [https://pathtocheckpoints](https://pathtocheckpoints) for more information.
    """,
)

train_group.command(
    name="text-to-spec",
    short_help="Train your Text-to-Spec model",
    help=f"""Train your text-to-spec model.  For example:

    **everyvoice train text-to-spec config/{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}.yaml**
    """,
)(train_fs2)

train_group.command(
    name="spec-to-wav",
    short_help="Train your Spec-to-Wav model",
    help=f"""Train your spec-to-wav model. For example:

    **everyvoice train spec-to-wav config/{SPEC_TO_WAV_CONFIG_FILENAME_PREFIX}.yaml**
    """,
)(train_hfg)

app.add_typer(
    train_group,
    name="train",
    short_help="Train your EveryVoice models",
)

# Add synthesize commands
synthesize_group = typer.Typer(
    pretty_exceptions_show_locals=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    rich_markup_mode="markdown",
    cls=TyperGroupOrderAsDeclared,
    help="""
    # Synthesize Help

        - **from-text** --- This is the most common input for performing normal speech synthesis. It will take text or a filelist with text and produce either waveform audio or spectrogram.

        - **from-spec** --- This is the model that turns your spectral features into audio. This type of synthesis is also known as copy synthesis and unless you know what you are doing, you probably don't want to do this.
    """,
)

synthesize_group.command(
    name="from-text",
)(synthesize_fs2)

synthesize_group.command(
    name="from-spec",
)(synthesize_hfg)

app.add_typer(
    synthesize_group,
    name="synthesize",
    short_help="Synthesize using your pre-trained EveryVoice models",
)

app.command(
    name="inspect-checkpoint",
    short_help="Extract structural information from a checkpoint",
)(inspect_checkpoint)


class TestSuites(str, Enum):
    all = "all"
    cli = "cli"
    config = "config"
    dev = "dev"
    model = "model"
    preprocessing = "preprocessing"
    text = "text"
    fs2 = "fs2"


@app.command(hidden=True)
def test(suite: TestSuites = typer.Argument(TestSuites.dev)):
    """Run a test suite"""
    from everyvoice.run_tests import run_tests

    run_tests(suite)


# Deferred full initialization to optimize the CLI, but still exposed for unit testing.
SCHEMAS_TO_OUTPUT: dict[str, Any] = {}  # dict[str, type[BaseModel]]


@app.command()
def demo(
    text_to_spec_model: Path = typer.Argument(
        ...,
        file_okay=True,
        exists=True,
        dir_okay=False,
        help="The path to a trained text-to-spec EveryVoice model.",
        autocompletion=complete_path,
    ),
    spec_to_wav_model: Path = typer.Argument(
        ...,
        help="The path to a trained vocoder.",
        dir_okay=False,
        file_okay=True,
        autocompletion=complete_path,
    ),
    output_dir: Path = typer.Option(
        "synthesis_output",
        "--output-dir",
        "-o",
        file_okay=False,
        dir_okay=True,
        help="The directory where your synthesized audio should be written",
        autocompletion=complete_path,
    ),
    accelerator: str = typer.Option("auto", "--accelerator", "-a"),
):
    from everyvoice.demo.app import create_demo_app

    demo = create_demo_app(
        text_to_spec_model_path=text_to_spec_model,
        spec_to_wav_model_path=spec_to_wav_model,
        output_dir=output_dir,
        accelerator=accelerator,
    )
    demo.launch()


@app.command(hidden=True)
def update_schemas(
    out_dir: Path = typer.Option(
        None,
        "-o",
        "--out-dir",
        file_okay=False,
        dir_okay=True,
        exists=True,
        autocompletion=complete_path,
    ),
):
    """Update the JSON Schemas. This is hidden because you shouldn't be calling this unless you are
    a developer for EveryVoice. Note: Pydantic will raise some Warnings related to the Callable fields
    having string Schemas. These can be ignored.
    """
    if out_dir is None:
        schema_dir_path = Path(__file__).parent / ".schema"
    else:
        schema_dir_path = out_dir

    # Defer somewhat slow imports to optimize CLI
    from everyvoice.config.preprocessing_config import PreprocessingConfig
    from everyvoice.config.text_config import TextConfig
    from everyvoice.model.aligner.config import AlignerConfig
    from everyvoice.model.e2e.config import EveryVoiceConfig
    from everyvoice.model.feature_prediction.config import FeaturePredictionConfig
    from everyvoice.model.vocoder.config import VocoderConfig

    # We should not be changing the schema for patches, so only include major/minor version
    MAJOR_MINOR_VERSION = ".".join(VERSION.split(".")[:2])

    SCHEMAS_TO_OUTPUT.update(
        {
            f"{ALIGNER_CONFIG_FILENAME_PREFIX}-schema-{MAJOR_MINOR_VERSION}.json": AlignerConfig,
            f"{TEXT_TO_WAV_CONFIG_FILENAME_PREFIX}-schema-{MAJOR_MINOR_VERSION}.json": EveryVoiceConfig,
            f"{TEXT_TO_SPEC_CONFIG_FILENAME_PREFIX}-schema-{MAJOR_MINOR_VERSION}.json": FeaturePredictionConfig,
            f"{PREPROCESSING_CONFIG_FILENAME_PREFIX}-schema-{MAJOR_MINOR_VERSION}.json": PreprocessingConfig,
            f"{TEXT_CONFIG_FILENAME_PREFIX}-schema-{MAJOR_MINOR_VERSION}.json": TextConfig,
            f"{SPEC_TO_WAV_CONFIG_FILENAME_PREFIX}-schema-{MAJOR_MINOR_VERSION}.json": VocoderConfig,
        }
    )

    for filename, schema in SCHEMAS_TO_OUTPUT.items():
        if (schema_dir_path / filename).exists():
            raise FileExistsError(
                f"Sorry a schema already exists for version {filename}.\n"
                "If it's already been published to the schema store, please bump the EveryVoice minor version number and generate the schemas again.\n"
                "If the current minor version is still in development, just delete the schema files and try again."
            )
        with open(schema_dir_path / filename, "w") as f:
            json.dump(schema.model_json_schema(), f, indent=2)
            f.write("\n")


CLICK_APP = typer.main.get_group(app)

if __name__ == "__main__":
    app()



class InvalidConfiguration(Exception):
    def __init__(self, msg):
        super().__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg


class ConfigError(Exception):
    pass


class OutOfVocabularySymbolError(Exception):
    pass


class BadDataError(Exception):
    pass



from pathlib import Path


def validate_path(value: str, is_dir=False, is_file=False, exists=False):
    path = Path(value).expanduser()
    if bool(is_dir) == bool(is_file):
        raise ValueError(
            "The path must be either a file or directory, but both or neither were specified."
        )
    if is_dir and path.is_file():
        print(f"Sorry, the path '{path}' is a file. Please select a directory.")
        return False
    if is_file and path.is_dir():
        print(f"Sorry, the path '{path}' is a directory. Please select a file.")
        return False
    if not exists and path.exists():
        print(f"Sorry, the path '{path}' already exists. Please choose another path.")
        return False
    if exists and not path.exists():
        print(f"Sorry, the path '{path}' doesn't exist. Please choose another path.")
        return False
    return True



import os
import sys

# Here, we add git submodule to path to allow imports to work as expected (i.e. from everyvoice.model.aligner import DeepForcedAligner)
# This is similar to dynamically adding a __init__.py at the root of the submodule, which don't do for the following reasons:
#   - This would cause this error with mypy: https://github.com/python/mypy/issues/8944
#   - This would cause this error with setuptools and imports from within the submodule.
#     I.e. we want to be able to import like: from dfaligner.cli import app
# New submodules for other models should follow this pattern and be added here
(parent_folder_path, current_dir) = os.path.split(os.path.dirname(__file__))

sys.path.append(
    os.path.join(parent_folder_path, "model", "aligner", "DeepForcedAligner")
)
sys.path.append(os.path.join(parent_folder_path, "model", "aligner", "wav2vec2aligner"))
sys.path.append(
    os.path.join(
        parent_folder_path, "model", "feature_prediction", "FastSpeech2_lightning"
    )
)
sys.path.append(
    os.path.join(parent_folder_path, "model", "vocoder", "HiFiGAN_iSTFT_lightning")
)



import pickle
from pathlib import Path
from typing import Any

import yaml
from pympi.Praat import TextGrid

from .duration_extraction import extract_durations_beam, extract_durations_with_dijkstra


def create_textgrid(save_path, tokens, durations, hop_size, sample_rate):
    tg = TextGrid(xmax=(sum(durations) * hop_size) / sample_rate)
    token_tier = tg.add_tier(name="Tokens")
    current_time = 0
    for i, d in enumerate(durations):
        dur_seconds = (d * hop_size) / sample_rate
        token_tier.add_interval(
            begin=current_time, end=current_time + dur_seconds, value=tokens[i]
        )
        current_time += dur_seconds
    tg.to_file(save_path)


def read_metafile(path: str) -> dict[str, str]:
    text_dict = {}
    with open(path, encoding="utf-8") as f:
        for line in f:
            split = line.split("|")
            text_id, text = split[0], split[-1]
            text_dict[text_id] = text.strip()
    return text_dict


def read_config(path: str) -> dict[str, Any]:
    with open(path, "r") as stream:
        config = yaml.load(stream, Loader=yaml.FullLoader)
    return config


def save_config(config: dict[str, Any], path: str) -> None:
    with open(path, "w+", encoding="utf-8") as stream:
        yaml.dump(config, stream, default_flow_style=False)


def get_files(path: str, extension=".wav") -> list[Path]:
    return list(Path(path).expanduser().resolve().rglob(f"*{extension}"))


def pickle_binary(data: object, file: str | Path) -> None:
    with open(str(file), "wb") as f:
        pickle.dump(data, f)


def unpickle_binary(file: str | Path) -> Any:
    with open(str(file), "rb") as f:
        return pickle.load(f)


def extract_durations_for_item(item, tokens, pred, method: str = "beam"):
    tokens_len, mel_len = item["tokens_len"], item["mel_len"]
    tokens = tokens[:tokens_len]
    pred = pred[:mel_len, :]
    match method:
        case "beam":
            duration_candidates, _ = extract_durations_beam(tokens, pred, 10)
            durations = duration_candidates[0]
        case "dijkstra":
            durations = extract_durations_with_dijkstra(tokens, pred)
        case _:
            raise NotImplementedError(f"Sorry, method '{method}' is not implemented")

    return item, durations



"""
This file is for types that are reused, but potentionally expensive to load.
Types requiring pydantic belong here.
Enums and other light-weight types should be defined in type_definitions.py instead.

Be careful not to cause this file to be imported when the CLI is launched, so that
everyvoice -h and command line completion can remain fast.
"""

from pydantic import BaseModel, ConfigDict


class InferenceControl(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    pitch: float = 1.0
    energy: float = 1.0
    duration: float = 1.0


class StatsInfo(BaseModel):
    min: float
    max: float
    std: float
    mean: float
    norm_min: float
    norm_max: float


class Stats(BaseModel):
    pitch: StatsInfo
    energy: StatsInfo
